#!/usr/bin/env python
# Copyright (c) 2011-13 Walter Bender

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

help_palettes = {}
help_windows = {}
palette_names = []
palettes = {}
palette_i18n_names = []
palette_init_on_start = []
palette_blocks = []
block_colors = []
expandable_blocks = []
block_names = {}
block_primitives = {}
default_values = {}
logo_commands = {}
logo_functions = {}
special_names = {}  # Names for blocks without names for popup help
content_blocks = ['number', 'string', 'description', 'audio', 'video',
                  'journal']
hidden_proto_blocks = []  # proto blocks that are (at least initially) hidden
value_blocks = []  # blocks whose labels are updated get added here
special_block_colors = {}
string_or_number_args = []
block_styles = {'basic-style': [],
                'blank-style': [],
                'basic-style-head': [],
                'basic-style-head-1arg': [],
                'basic-style-tail': [],
                'basic-style-extended': [],
                'basic-style-extended-vertical': [],
                'basic-style-1arg': [],
                'basic-style-2arg': [],
                'basic-style-3arg': [],
                'basic-style-7arg': [],
                'basic-style-var-arg': [],
                'bullet-style': [],
                'box-style': [],
                'box-style-media': [],
                'number-style': [],
                'number-style-var-arg': [],
                'number-style-var-3arg': [],
                'number-style-block': [],
                'number-style-porch': [],
                'number-style-1arg': [],
                'number-style-1strarg': [],
                'compare-style': [],
                'compare-porch-style': [],
                'boolean-style': [],
                'not-style': [],
                'boolean-block-style': [],
                'boolean-1arg-block-style': [],
                'flow-style-tail': [],
                'clamp-style': [],
                'clamp-style-collapsible': [],
                'clamp-style-collapsed': [],
                'clamp-style-1arg': [],
                'clamp-style-hat-1arg': [],
                'clamp-style-hat': [],
                'clamp-style-boolean': [],
                'clamp-style-until': [],
                'clamp-style-else': [],
                'portfolio-style-2x2': [],
                'portfolio-style-1x1': [],
                'portfolio-style-2x1': [],
                'portfolio-style-1x2': []}


from gi.repository import Gtk
from gi.repository import Gdk

try:
    from sugar3.graphics import style
    from .util.helpbutton import (add_section, add_paragraph)
    GRID_CELL_SIZE = style.GRID_CELL_SIZE
    HELP_PALETTE = True
except ImportError:
    GRID_CELL_SIZE = 55
    HELP_PALETTE = False

from .taconstants import (EXPANDABLE_STYLE, EXPANDABLE_FLOW)

from gettext import gettext as _

help_strings = {
    'next': _('displays next palette'),
    'orientation': _("changes the orientation of the palette of blocks")}


class Palette():

    """ a class for defining new palettes """

    def __init__(self, name, colors=["#00FF00", "#00A000"], position=None):
        self._name = name
        self._special_name = _(name)
        self._colors = colors
        self._max_text_width = int(Gdk.Screen.width() / 3) - 20

        # Prepare a vbox for the help palette
        if self._name not in help_palettes:
            self._help_box = Gtk.VBox()
            self._help_box.set_homogeneous(False)
            help_palettes[self._name] = self._help_box
            help_windows[self._name] = Gtk.ScrolledWindow()
            help_windows[self._name].set_size_request(
                int(Gdk.Screen.width() / 3),
                Gdk.Screen.height() - GRID_CELL_SIZE * 3)
            help_windows[self._name].set_policy(
                Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
            help_windows[self._name].add_with_viewport(
                help_palettes[self._name])
            help_palettes[self._name].show()
            self._help = None
        else:
            self._help_box = help_palettes[self._name]
            self._help = 'deja vu'

    def add_palette(self, position=None, init_on_start=False):
        if self._name is None:
            print('You must specify a name for your palette')
            return

        # Insert new palette just before the trash
        if 'trash' in palette_names:
            i = palette_names.index('trash')
        elif _('trash') in palette_i18n_names:
            i = palette_i18n_names.index(_('trash'))
        else:
            i = len(palette_names)

        if position is not None and isinstance(position, int) and position < i:
            i = position

        if self._name not in palette_names:
            palettes[self._name] = self
            palette_names.insert(i, self._name)
            palette_i18n_names.insert(i, _(self._name))
            palette_blocks.insert(i, [])
            block_colors.insert(i, self._colors)
            if init_on_start:
                if self._name not in palette_init_on_start:
                    palette_init_on_start.append(self._name)
        else:
            return

        # Special name entry is needed for help hover mechanism
        special_names[self._name] = self._special_name
        if self._help is not None:
            help_strings[self._name] = self._help
        else:
            help_strings[self._name] = ''

    def set_help(self, help):
        if self._help is None:
            self._help = help
            if hasattr(self, '_help_box') and HELP_PALETTE:
                add_section(self._help_box, self._help,
                            icon=self._name + 'off')

    def set_special_name(self, name):
        self._special_name = name

    def add_block(self, block_name, style='basic-block', label=None,
                  special_name=None, default=None, prim_name=None,
                  help_string=None, value_block=False, content_block=False,
                  logo_command=None, hidden=False, colors=None,
                  string_or_number=False, before=None, after=None,
                  private=None):
        """ Add a new block to the palette """
        block = _ProtoBlock(block_name)
        block.set_style(style)
        if label is not None:
            block.set_label(label)
        if special_name is not None:
            block.set_special_name(special_name)
        if default is not None:
            if default == 'None':
                block.set_default(None)
            else:
                block.set_default(default)
        if prim_name is not None:
            block.set_prim_name(prim_name)
        if logo_command is not None:
            block.set_logo_command(logo_command)
        if help_string is not None:
            block.set_help(help_string)
            if not hidden:
                first_arg = None
                if special_name is None:
                    if isinstance(label, list):
                        first_arg = label[0]
                    else:
                        first_arg = label
                else:
                    first_arg = special_name
                if HELP_PALETTE:
                    if first_arg is None or \
                            first_arg == '' or \
                            first_arg == ' ':
                        add_paragraph(self._help_box, '%s' % (help_string))
                    else:
                        add_paragraph(self._help_box, '%s: %s' % (first_arg,
                                                                  help_string))
        if colors is not None:
            block.set_colors(colors)
        if string_or_number:
            block.set_string_or_number()
        if before is not None:
            block.before = before
        if after is not None:
            block.after = after
        if private is not None:
            block.private = private
        block.set_value_block(value_block)
        block.set_content_block(content_block)
        block.set_palette(self._name)
        if hidden:
            block.set_hidden()
        block.add_block()


def make_palette(palette_name, colors=None, help_string=None, position=None,
                 init_on_start=False, translation=None):
    """ Palette helper function """
    if palette_name in palettes:
        return palettes[palette_name]
    else:
        if colors is None:
            palette = Palette(palette_name)
        else:
            palette = Palette(palette_name, colors)
        if help_string is not None:
            palette.set_help(help_string)
        palette.add_palette(position, init_on_start=init_on_start)
        return palette


def palette_name_to_index(palette_name):
    ''' Find the index associated with palette_name. '''
    if palette_name in palette_names:
        return palette_names.index(palette_name)
    elif palette_name in palette_i18n_names:
        return palette_i18n_names.index(palette_name)
    else:
        return None


def define_logo_function(key, value):
    ''' Add a logo function to the table (not necessarily associated
    with a block, e.g., color lookup tables) '''
    logo_functions[key] = value


class _ProtoBlock():

    """ a class for defining new block primitives """

    def __init__(self, name):
        self._name = name
        self._special_name = None
        self._palette = None
        self._style = None
        self._label = None
        self._default = None
        self._help = None
        self._prim_name = None
        self._logo_command = None
        self._value_block = False
        self._content_block = False
        self._colors = None
        self._hidden = False
        self._string_or_number = False

    def add_block(self, position=None):
        if self._name is None:
            print('You must specify a name for your block')
            return

        # FIXME: Does the block already exist? A block can live on
        # multiple palettes, but it can only have one set of
        # atttributes. So if this is a redefinition, remove it from
        # all lists except palettes before regeneration.

        if self._style is None:
            print('You must specify a style for your block')
            return
        else:
            block_styles[self._style].append(self._name)
        if self._style in ['clamp-style',
                           'clamp-style-hat-1arg',
                           'clamp-style-hat',
                           'clamp-style-collapsible',
                           'clamp-style-1arg',
                           'clamp-style-boolean',
                           'clamp-style-until',
                           'clamp-style-else']:
            EXPANDABLE_FLOW.append(self._name)

        if self._label is not None:
            block_names[self._name] = self._label

        if self._palette is not None:
            i = palette_names.index(self._palette)
            if self._name in palette_blocks[i]:
                print('%s already in palette %s, skipping...' % \
                    (self._name, self._palette))
            else:
                if position is not None and isinstance(position, int) and \
                        position < len(palette_blocks[i]):
                    palette_blocks[i].insert(position, self._name)
                else:
                    palette_blocks[i].append(self._name)
                    if position is not None:
                        print('Ignoring position (%s)' % (str(position)))

        if self._help is not None:
            help_strings[self._name] = self._help
        else:
            help_strings[self._name] = ''

        if self._value_block:
            value_blocks.append(self._name)

        if self._content_block:
            content_blocks.append(self._name)

        if self._prim_name is not None:
            block_primitives[self._name] = self._prim_name

        if self._logo_command is not None and self._prim_name is not None:
            logo_commands[self._prim_name] = self._logo_command

        if self._default is not None:
            default_values[self._name] = self._default

        if self._special_name is not None:
            special_names[self._name] = self._special_name

        if self._style in EXPANDABLE_STYLE:
            expandable_blocks.append(self._name)

        if self._colors is not None:
            special_block_colors[self._name] = self._colors

        if self._string_or_number:
            string_or_number_args.append(self._name)

        if self._hidden:
            hidden_proto_blocks.append(self._name)

    def set_hidden(self):
        self._hidden = True

    def set_colors(self, colors=None):
        self._colors = colors

    def set_string_or_number(self, flag=True):
        self._string_or_number = flag

    def set_value_block(self, value=True):
        self._value_block = value

    def set_content_block(self, value=True):
        self._content_block = value

    def set_palette(self, palette):
        if palette not in palette_names:
            print('Could not find palette %s' % (palette))
        else:
            self._palette = palette

    def set_help(self, help):
        self._help = help

    def set_special_name(self, name):
        self._special_name = name

    def set_label(self, label):
        if isinstance(label, list):
            self._label = label[:]
        else:
            self._label = [label]

    def set_default(self, default):
        if isinstance(default, list):
            self._default = default[:]
        else:
            self._default = [default]

    def set_style(self, style):
        if style not in block_styles:
            print('Unknown style: %s' % (style))
        else:
            self._style = style

    def set_prim_name(self, prim_name):
        self._prim_name = prim_name

    def set_logo_command(self, logo_command):
        self._logo_command = logo_command



#!/usr/bin/env python

import sys
from configparser import ConfigParser
import xml.etree.cElementTree as ET


def get_files(path):
    files = []
    for name in os.listdir(path):
        files.append(os.path.join(path, name))
    return files


def generate_appdata(prefix, bundle_id):
    info = ConfigParser()
    info.read(os.path.join('activity', 'activity.info'))

    required_fields = ['metadata_license', 'license', 'name', 'icon',
                       'description']
    for name in required_fields:
        if not info.has_option('Activity', name):
            print('[WARNING] Activity needs more metadata for AppStream '
                  'file')
            print('  Without an AppStream file, the activity will NOT '
                  'show in software stores!')
            print('  Please `pydoc sugar3.activity.bundlebuilder` for'
                  'more info')
            return

    # See https://www.freedesktop.org/software/appstream/docs/
    root = ET.Element('component', type='desktop')
    ET.SubElement(root, 'translation', type='gettext').text = \
        bundle_id
    ET.SubElement(root, 'id').text = \
        bundle_id + '.desktop'
    desc = ET.fromstring('<description>{}</description>'.format(
        info.get('Activity', 'description')))
    root.append(desc)

    ET.SubElement(root, 'content_rating', type='oars-1.1')

    release_pairs = [('218', '2018-05-22'),
                     ('216', '2017-06-24'),
                     ('215', '2017-04-19')]
    releases_root = ET.SubElement(root, 'releases')
    for version, date in release_pairs:
        ET.SubElement(releases_root, 'release', date=date, version=version)

    copy_pairs = [('metadata_license', 'metadata_license'),
                  ('license', 'project_license'),
                  ('summary', 'summary'),
                  ('name', 'name')]
    for key, ename in copy_pairs:
        ET.SubElement(root, ename).text = info.get('Activity', key)

    if info.has_option('Activity', 'screenshots'):
        screenshots = info.get('Activity', 'screenshots').split()
        ss_root = ET.SubElement(root, 'screenshots')
        for i, screenshot in enumerate(screenshots):
            e = ET.SubElement(ss_root, 'screenshot')
            if i == 0:
                e.set('type', 'default')
            ET.SubElement(e, 'image').text = screenshot

    if info.has_option('Activity', 'url'):
        ET.SubElement(root, 'url', type='homepage').text = \
            info.get('Activity', 'url')
    if info.has_option('Activity', 'repository_url'):
        ET.SubElement(root, 'url', type='bugtracker').text = \
            info.get('Activity', 'repository_url')
    elif info.has_option('Activity', 'repository'):
        ET.SubElement(root, 'url', type='bugtracker').text = \
            info.get('Activity', 'repository')

    path = os.path.join(prefix, 'share', 'metainfo',
                        bundle_id + '.appdata.xml')
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    tree = ET.ElementTree(root)
    tree.write(path, encoding='UTF-8')


if len(sys.argv) > 1 and '--no-sugar' == sys.argv[1]:
    # Remove the argument from the stack so we don't cause problems
    # for distutils
    sys.argv.pop(1)

    import os
    import shutil
    import subprocess
    from distutils.core import setup
    from distutils.command.install import install

    class post_install(install):
        def run(self):
            install.run(self)
            print("Running post_install")

            generate_appdata(self.prefix,
                             'org.laptop.TurtleArtActivity')

            # Create a simple module that allows the app where to discover
            # its lib and share content.
            dst = os.path.join(self.install_purelib, "TurtleArt")
            fd = open(os.path.join(dst, "installinfo.py"), "w")
            fd.write("INSTALL_PREFIX='%s'\n" % self.prefix)
            fd.close()

            root = self.root or ''

            # distutils doesn't offer a nice way to do recursive install of
            # a directory tree, so we install the remaining parts here.
            libdir = os.path.join(root + self.install_base, "lib",
                                  "TurtleBlocks")
            if not os.path.isdir(libdir):
                os.makedirs(libdir)

            sharedir = os.path.join(root + self.install_base, "share",
                                    "TurtleBlocks")
            if not os.path.isdir(sharedir):
                os.makedirs(sharedir)

            shutil.copytree("plugins", os.path.join(libdir, "plugins"))
            shutil.copytree("samples", os.path.join(sharedir, "samples"))

            localedir = os.path.join(root + self.install_base, "share",
                                     "locale")
            for f in os.listdir('po'):
                if not f.endswith('.po') or f == 'pseudo.po':
                    continue

                file_name = os.path.join('po', f)
                lang = f[:-3]

                mo_path = os.path.join(localedir, lang, 'LC_MESSAGES')
                if not os.path.isdir(mo_path):
                    os.makedirs(mo_path)

                mo_file = os.path.join(
                    mo_path, 'org.laptop.TurtleArtActivity.mo')
                retcode = subprocess.call(
                    ['msgfmt', '--output-file=%s' % mo_file, file_name])
                if retcode:
                    print('ERROR - msgfmt failed with return code %i.'
                          % retcode)

    DATA_FILES = [
        ('activity', get_files('activity/')),
        ('icons', get_files('icons/')),
        ('images', get_files('images/')),
        ('/usr/share/applications', ['turtleblocks.desktop']),
        ('org.laptop.TurtleArtActivity.gschema.xml')
    ]

    setup(name='Turtle Art',
          description="A LOGO-like tool for teaching programming",
          author="Walter Bender",
          author_email="walter.bender@gmail.com",
          version='0.9.4',
          packages=['TurtleArt', 'TurtleArt.util'],
          scripts=['turtleblocks'],
          data_files=DATA_FILES,
          cmdclass={"install": post_install}
          )
else:
    from sugar3.activity import bundlebuilder

    if __name__ == "__main__":
        bundlebuilder.start()



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Walter Bender

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import cairo

from gi.repository import Gdk
from gi.repository import GdkPixbuf

from .taconstants import EXPANDABLE, EXPANDABLE_ARGS, OLD_NAMES, CONSTANTS, \
    STANDARD_STROKE_WIDTH, BLOCK_SCALE, BOX_COLORS, \
    GRADIENT_COLOR, EXPANDABLE_FLOW, Color, \
    MEDIA_BLOCK2TYPE, BLOCKS_WITH_SKIN

from .tapalette import palette_blocks, block_colors, expandable_blocks, \
    content_blocks, block_names, block_primitives, \
    block_styles, special_block_colors

from .tasprite_factory import SVG, svg_str_to_pixbuf
from . import sprites

from .tautils import debug_output, error_output


media_blocks_dictionary = {}  # new media blocks get added here


class Media(object):

    """ Media objects can be images, audio files, videos, Journal
    descriptions, or camera snapshots. """

    ALL_TYPES = ('media', 'audio', 'video', 'descr', 'camera', 'camera1')

    def __init__(self, media_type, value=None):
        """
        media_type --- a string that indicates the kind of media:
            media --- image
            audio --- audio file
            video --- video
            descr --- Journal description
            camera, camera1 --- camera snapshot
        value --- a file path or a reference to a Sugar datastore object """
        if media_type == 'image':
            media_type = 'media'
        if media_type not in Media.ALL_TYPES:
            raise ValueError("Media.type must be one of " +
                             repr(Media.ALL_TYPES))
        self.type = media_type
        self.value = value

    def __str__(self):
        return '%s_%s' % (self.type, str(self.value))

    def __repr__(self):
        return 'Media(type=%s, value=%s)' % (repr(self.type), repr(self.value))


class Blocks:

    """ A class for the list of blocks and everything they share in common """

    def __init__(self, font_scale_factor=1, decimal_point='.'):
        self.list = []
        self.max_width = 400
        self.font_scale_factor = font_scale_factor
        self.decimal_point = decimal_point

    def get_block(self, i):
        if i < 0 or i > len(self.list) - 1:
            return(None)
        else:
            return(self.list[i])

    def swap(self, blk1, blk2):
        i1 = self.list.index(blk1)
        i2 = self.list.index(blk2)
        self.list[i1] = blk2
        self.list[i2] = blk1

    def length_of_list(self):
        return(len(self.list))

    def append_to_list(self, block):
        self.list.append(block)

    def remove_from_list(self, block):
        if block in self.list:
            self.list.remove(block)

    def print_list(self, block_type=None):
        for i, block in enumerate(self.list):
            if block_type is None or block_type == block.type:
                print("%d: %s" % (i, block.name))

    def set_scale(self, scale):
        for b in self.list:
            for i in range(len(b.font_size)):
                b.font_size[i] *= b.scale * scale / self.font_scale_factor
        self.font_scale_factor = scale

    def spr_to_block(self, spr):
        for b in self.list:
            if spr == b.spr:
                return b
        return None

    def get_next_block(self, block):
        if block is None:
            return None
        try:
            i = self.list.index(block)
        except ValueError:
            return None
        i += 1
        if i < len(self.list):
            return self.list[i]
        else:
            return self.list[0]

    def get_next_block_of_same_type(self, block):
        if block is None:
            return None
        type = block.type
        i = 0
        while block is not None:
            block = self.get_next_block(block)
            if block is not None:
                if block.type == type:
                    return block
            if i == len(self.list):
                break
            i += 1
        return None

    def get_similar_blocks(self, block_type, name):
        block_list = []
        if isinstance(name, str):
            for block in self.list:
                if block.type == block_type and block.name == name:
                    block_list.append(block)
        else:
            for block in self.list:
                if block.type == block_type and block.name in name:
                    block_list.append(block)
        return block_list


class Block:

    """ A class for the individual blocks

    Attributes:
    docks -- a list of docks, i.e. connection points where other blocks
        could be attached. Each dock is a list of the form
        [type_of_dock, flow_is_in, x, y, parenthesis]
        with the last element being optional.
        type_of_dock may be one of the following strings:
            flow -- connect to the previous or next block ('slot' or 'tab')
            bool, media, number, string -- argument slot ('innie') or
                return value ('outie') of the given kind
            unavailable -- nothing can be attached here ('cap' or 'tail')
        flow_is_in is True if the flow is into the block, or False for out.
        x and y are coodinates for positioning the block on the dock.
        parenthesis is only used with arguments and ensures a known order
            of arguments for arithmetic and logical operations.
    connections -- a list of blocks that are attached to this one (or that
        this one is attached to). This list corresponds to the docks list
        as it uses the same indices. Slots where nothing is attached are
        None on this list.
    primitive -- a callable that is called when the block is executed
    type -- type of the block:
        block -- block that is part of the user's program
        proto -- block on a palette, used to generate other blocks
        trash -- block in the trash """

    def __init__(self, block_list, sprite_list, name, x, y, type='block',
                 values=None, scale=BLOCK_SCALE[3],
                 colors=['#A0A0A0', '#808080']):

        self.block_list = block_list
        self.spr = None
        self.shapes = [None, None]
        self.name = name
        self.colors = colors
        self._custom_colors = False
        self.scale = scale
        self.docks = None
        self.connections = None
        self.status = None
        self.values = []
        self.primitive = None
        self.type = type
        self.dx = 0
        self.ex = 0
        self.ey = 0
        self.ey2 = 0
        self._ei = 0
        self.font_size = [6.0, 4.5]
        self._image = None
        self._visible = True
        self.unknown = False  # Block is of unknown style
        # Private method called before a block instance is run
        self.before = None
        # Private method called after a block instance is run
        self.after = None
        self.private = None  # Private data for block primitive

        self._block_methods = {
            'basic-style': self._make_basic_style,
            'blank-style': self._make_blank_style,
            'basic-style-head': self._make_basic_style_head,
            'basic-style-head-1arg': self._make_basic_style_head_1arg,
            'basic-style-tail': self._make_basic_style_tail,
            'basic-style-extended': [self._make_basic_style, 16, 16],
            'basic-style-extended-vertical': [self._make_basic_style, 0, 4],
            'basic-style-1arg': self._make_basic_style_1arg,
            'basic-style-2arg': self._make_basic_style_2arg,
            'basic-style-3arg': self._make_basic_style_3arg,
            'basic-style-7arg': self._make_basic_style_7arg,
            'basic-style-var-arg': self._make_basic_style_var_arg,
            'bullet-style': self._make_bullet_style,
            'box-style': self._make_box_style,
            'box-style-media': self._make_media_style,
            'number-style': self._make_number_style,
            'number-style-block': self._make_number_style_block,
            'number-style-porch': self._make_number_style_porch,
            'number-style-1arg': self._make_number_style_1arg,
            'number-style-1strarg': self._make_number_style_1strarg,
            'number-style-var-arg': self._make_number_style_var_arg,
            'number-style-var-3arg': self._make_number_style_var_3arg,
            'compare-style': self._make_compare_style,
            'compare-porch-style': self._make_compare_porch_style,
            'boolean-style': self._make_boolean_style,
            'not-style': self._make_not_style,
            'boolean-block-style': self._make_boolean_block_style,
            'boolean-1arg-block-style': self._make_boolean_1arg_block_style,
            'clamp-style': self._make_clamp_style,
            'clamp-style-collapsible': self._make_clamp_style_collapsible,
            'clamp-style-collapsed': self._make_clamp_style_collapsed,
            'clamp-style-1arg': self._make_clamp_style_1arg,
            'clamp-style-hat-1arg': self._make_clamp_style_hat_1arg,
            'clamp-style-hat': self._make_clamp_style_hat,
            'clamp-style-boolean': self._make_clamp_style_boolean,
            'clamp-style-until': self._make_clamp_style_until,
            'clamp-style-else': self._make_clamp_style_else,
            'flow-style-tail': self._make_flow_style_tail,
            'portfolio-style-2x2': self._make_portfolio_style_2x2,
            'portfolio-style-1x1': self._make_portfolio_style_1x1,
            'portfolio-style-2x1': self._make_portfolio_style_2x1,
            'portfolio-style-1x2': self._make_portfolio_style_1x2}

        if self.name in OLD_NAMES:
            self.name = OLD_NAMES[self.name]

        for i in range(len(self.font_size)):
            self.font_size[i] *= self.scale * \
                self.block_list.font_scale_factor

        if values is not None:
            for v in (values):
                self.values.append(v)

        # If there is already a block with the same name, reuse it
        copy_block = None
        if self.cloneable():
            for b in self.block_list.list:
                if b.scale == self.scale and b.name == self.name:
                    copy_block = b
                    break
        self._new_block_from_factory(sprite_list, x, y, copy_block)

        if name in block_primitives:
            self.primitive = block_primitives[self.name]

        self.block_list.append_to_list(self)

    def __repr__(self):
        if self.is_value_block():
            name = self.get_value()
        else:
            name = self.name
        return 'Block(%s)' % (repr(name))

    def get_visibility(self):
        ''' Should block be visible on the palette? '''
        return self._visible

    def set_visibility(self, state):
        ''' Should block be visible? '''
        self._visible = state
        if self._visible:
            self.spr.restore()
        else:
            self.spr.hide()

    def expandable(self):
        """ Can this block be expanded? """
        if self.name in EXPANDABLE:
            return True
        if self.name in expandable_blocks:
            return True
        if self.name in EXPANDABLE_ARGS:
            return True
        if self.name in EXPANDABLE_FLOW:
            return True
        return False

    def cloneable(self):
        """ Is it safe to clone this block? """
        if self.expandable():
            return False
        if self.name in block_styles['box-style']:
            return False
        if self.name in ['storein', 'box', 'string',
                         # Deprecated blocks
                         'sandwichtop', 'sandwichtop_no_label']:
            return False
        return True

    def is_value_block(self):
        """ Return True iff this block is a value block (numeric, string,
        media, etc.) """
        return self.primitive is None and self.values

    def get_value(self, add_type_prefix=True):
        """ Return the value stored in this value block or None if this is
        not a value block
        add_type_prefix -- prepend a prefix to indicate the type of the
            'raw' value """
        if not self.is_value_block():
            return None

        if self.name == 'number':
            try:
                return float(self.values[0])
            except ValueError:
                return float(ord(self.values[0][0]))
        elif (self.name == 'string' or
                self.name == 'title'):  # deprecated block
            if add_type_prefix:
                result = '#s'
            else:
                result = ''
            if isinstance(self.values[0], (float, int)):
                if int(self.values[0]) == self.values[0]:
                    self.values[0] = int(self.values[0])
                result += str(self.values[0])
            else:
                result += self.values[0]
            return result
        elif self.name in MEDIA_BLOCK2TYPE:
            return Media(MEDIA_BLOCK2TYPE[self.name], self.values[0])
        elif self.name in media_blocks_dictionary:
            return Media('media', self.name.upper())
        else:
            return None

    def highlight(self):
        """ We may want to highlight a block... """
        if self.spr is not None and self.status is not 'collapsed':
            self.spr.set_shape(self.shapes[1])

    def unhighlight(self):
        """ Or unhighlight it. """
        if self.spr is not None and self.status is not 'collapsed':
            self.spr.set_shape(self.shapes[0])

    def resize(self):
        """ We need to resize some blocks on the fly so the labels fit. """
        if self.spr is None:
            return
        dx = (self.spr.label_width() - self.spr.label_safe_width()) / \
            self.scale
        if self.dx + dx >= self.block_list.max_width and self.name == 'string':
            self.dx = self.block_list.max_width
            self.refresh()
            for i, label in enumerate(self.spr.labels):
                self._set_labels(i, label)
        elif dx != 0:
            self.dx += dx
            if self.dx < 0:
                self.dx = 0
            self.refresh()

    def set_image(self, image, x, y):
        """ Some blocks get a skin. """
        if self.spr is None:
            return
        self._image = image
        self.spr.set_image(image, 1, x, y)

    def scale_image(self, x, y, w, h):
        """ The skin might need scaling. """
        if self.spr is None:
            return
        if self._image is not None:
            tmp = self._image.scale_simple(w, h, GdkPixbuf.InterpType.NEAREST)
            self.spr.set_image(tmp, 1, x, y)

    def rescale(self, scale):
        """ We may want to rescale blocks as well. """
        if self.spr is None:
            return
        for i in range(len(self.font_size)):
            self.font_size[i] /= self.scale
        self.scale = scale
        for i in range(len(self.font_size)):
            self.font_size[i] *= self.scale
        self.svg.set_scale(self.scale)
        self.refresh()
        self.spr.inval()

    def set_colors(self, colors):
        self.colors = colors[:]
        self._custom_colors = True
        self.refresh()

    def refresh(self):
        if self.spr is None:
            return
        self._make_block(self.svg)
        self._set_margins()
        self._set_label_attributes()
        self.spr.set_shape(self.shapes[0])

    def add_arg(self, keep_expanding=True):
        """ We may want to add additional slots for arguments ("innies"). """
        if self.spr is None:
            return
        h = self.svg.get_height()
        self._ei += 1
        if self.type == 'block' and keep_expanding:
            self.svg.set_show(True)
        else:
            self.svg.set_show(False)
        self.refresh()
        return self.svg.get_height() - h

    def expand_in_y(self, dy):
        """ We may want to grow a block vertically. """
        if self.spr is None:
            return
        self.ey += dy
        if self.type == 'block':
            if self.ey > 0:
                self.svg.set_hide(True)
            else:
                self.svg.set_hide(False)
            self.svg.set_show(True)
        else:
            self.svg.set_hide(False)
            self.svg.set_show(False)
        self.refresh()

    def expand_in_y2(self, dy):
        """ We may want to grow a block vertically. """
        if self.spr is None:
            return
        self.ey2 += dy
        if self.type == 'block':
            if self.ey2 > 0:
                self.svg.set_hide(True)
            else:
                self.svg.set_hide(False)
            self.svg.set_show(True)
        else:
            self.svg.set_hide(False)
            self.svg.set_show(False)
        self.refresh()

    def expand_in_x(self, dx):
        """ We may want to grow a block horizontally. """
        if self.spr is None:
            return
        self.ex += dx
        if self.type == 'block':
            self.svg.set_hide(True)
            self.svg.set_show(True)
        else:
            self.svg.set_hide(False)
            self.svg.set_show(False)
        self.refresh()

    def contract_in_y(self, dy):
        """ We may want to shrink a block veritcally. """
        if self.spr is None:
            return
        self.ey -= dy
        if self.ey < 0:
            self.ey = 0
        if self.type == 'block':
            if self.ey > 0:
                self.svg.set_hide(True)
            else:
                self.svg.set_hide(False)
            self.svg.set_show(True)
        else:
            self.svg.set_hide(False)
            self.svg.set_show(False)
        self.refresh()

    def contract_in_y2(self, dy):
        """ We may want to shrink a block veritcally. """
        if self.spr is None:
            return
        self.ey2 -= dy
        if self.ey2 < 0:
            self.ey2 = 0
        if self.type == 'block':
            if self.ey2 > 0:
                self.svg.set_hide(True)
            else:
                self.svg.set_hide(False)
            self.svg.set_show(True)
        else:
            self.svg.set_hide(False)
            self.svg.set_show(False)
        self.refresh()

    def contract_in_x(self, dx):
        """ We may want to shrink a block horizontally. """
        if self.spr is None:
            return
        self.ex -= dx
        if self.ex < 0:
            self.ex = 0
        if self.type == 'block':
            if self.ex > 0:
                self.svg.set_hide(True)
            else:
                self.svg.set_hide(False)
            self.svg.set_show(True)
        else:
            self.svg.set_hide(False)
            self.svg.set_show(False)
        self.refresh()

    def reset_x(self):
        if self.spr is None:
            return 0
        dx = -self.ex
        self.ex = 0
        self.svg.set_hide(False)
        if self.type == 'block':
            self.svg.set_show(True)
        else:
            self.svg.set_show(False)
        self.refresh()
        return dx

    def reset_y(self):
        if self.spr is None:
            return 0
        dy = -self.ey
        self.ey = 0
        self.svg.set_hide(False)
        if self.type == 'block':
            self.svg.set_show(True)
        else:  # 'proto'
            self.svg.set_show(False)
        self.refresh()
        return dy

    def reset_y2(self):
        if self.spr is None:
            return 0
        dy = -self.ey2
        self.ey2 = 0
        self.svg.set_hide(False)
        if self.type == 'block':
            self.svg.set_show(True)
        else:  # 'proto'
            self.svg.set_show(False)
        self.refresh()
        return dy

    def get_expand_x_y(self):
        if self.spr is None:
            return(0, 0)
        return (self.ex, self.ey, self.ey2)

    def _new_block_from_factory(self, sprite_list, x, y, copy_block=None):
        self.svg = SVG()
        self.svg.set_scale(self.scale)
        self.svg.set_innie([False])
        self.svg.set_outie(False)
        self.svg.set_tab(True)
        self.svg.set_slot(True)

        if copy_block is not None:
            self._left = copy_block._left
            self._top = copy_block._top
            self._right = copy_block._right
            self._bottom = copy_block._bottom
            self.dx = copy_block.dx
            self.ex = copy_block.ex
            self.ey = copy_block.ey
            self.width = copy_block.width
            self.height = copy_block.height
            self.shapes[0] = copy_block.shapes[0]
            if sprite_list is not None:
                self.spr = sprites.Sprite(sprite_list, x, y, self.shapes[0])
                self.spr._margins = copy_block.spr._margins[:]
            if len(copy_block.shapes) > 1:
                self.shapes[1] = copy_block.shapes[1]
            self.docks = copy_block.docks[:]
        else:
            if self.expandable() and self.type == 'block':
                self.svg.set_show(True)

            self._make_block(self.svg)

            if sprite_list is not None:
                self.spr = sprites.Sprite(sprite_list, x, y, self.shapes[0])
                self._set_margins()

        self._set_label_attributes()
        if (self.name == 'number' or self.name == 'string') and \
                len(self.values) > 0:
            for i, v in enumerate(self.values):
                if v is not None:
                    if self.name == 'number':
                        self._set_labels(
                            i,
                            str(v).replace(
                                '.',
                                self.block_list.decimal_point))
                    else:
                        self._set_labels(i, str(v))
        elif self.type == 'block' and self.name in CONSTANTS:
            if isinstance(CONSTANTS[self.name], Color):
                v = int(CONSTANTS[self.name])
            else:
                v = CONSTANTS[self.name]
            if self.name not in BLOCKS_WITH_SKIN:
                self._set_labels(0, block_names[self.name][0] + ' = ' + str(v))

        elif self.name in block_names:
            for i, n in enumerate(block_names[self.name]):
                self._set_labels(i, n)

        if copy_block is None and self.spr is not None:
            if self.spr.label_width() > self.spr.label_safe_width():
                self.resize()

    def _set_margins(self):
        if self.spr is None:
            return
        self.spr.set_margins(self.svg.margins[0], self.svg.margins[1],
                             self.svg.margins[2], self.svg.margins[3])

    def _set_label_attributes(self):
        if self.spr is None:
            return
        if self.name in content_blocks:
            n = len(self.values)
            if n == 0:
                n = 1  # Force a scale to be set, even if there is no value.
        else:
            n = 0
            if self.name in block_names:
                n = len(self.spr.labels)
            elif self.name not in BLOCKS_WITH_SKIN:
                debug_output('WARNING: unknown block name %s' % (self.name))
        for i in range(n):
            if i > 0:
                size = int(self.font_size[1] + 0.5)
            else:
                size = int(self.font_size[0] + 0.5)
            if self.name in block_styles['compare-porch-style']:
                self.spr.set_label_attributes(size, True, 'center', 'bottom',
                                              i=i)
            elif self.name in block_styles['clamp-style-hat']:
                self.spr.set_margins(t=10 * self.scale)
                self.spr.set_label_attributes(size, True, 'center', 'top',
                                              i=i)
            elif self.name in block_styles['clamp-style-hat-1arg']:
                self.spr.set_label_attributes(size, True, 'center', 'top',
                                              i=i)
            elif self.name in block_styles['number-style-porch']:
                self.spr.set_label_attributes(size, True, 'right', 'bottom',
                                              i=i)
            elif self.name in EXPANDABLE_FLOW:
                self._calc_moving_labels(i)
            elif self.name == 'string':
                self.spr.set_label_attributes(size, False, 'center', 'middle')
            elif i == 1:  # top
                self.spr.set_label_attributes(size, True, 'right', 'top', i=i)
            elif i > 0 and i == n - 1:  # bottom
                self.spr.set_label_attributes(size, True, 'right', 'bottom',
                                              i=i)
            elif i > 0 and n > 4:
                self.spr.set_label_attributes(
                    size, True, 'right',
                    y_pos=self.docks[i][3] - self.font_size[1], i=i)
            elif i > 0:
                self.spr.set_label_attributes(size, True, 'right', 'middle',
                                              i=i)
            else:
                self.spr.set_label_attributes(size, True, 'center', 'middle',
                                              i=i)

    def _calc_moving_labels(self, i):
        ''' Some labels move as blocks change shape/size '''
        if self.name in block_styles['clamp-style'] or \
           self.name in block_styles['clamp-style-collapsible']:
            y = int((self.docks[0][3] + self.docks[1][3]) / 3.3)
            self.spr.set_label_attributes(int(self.font_size[0] + 0.5),
                                          True, 'right', y_pos=y, i=0)
        elif self.name in block_styles['clamp-style-1arg']:
            y = self.docks[1][3] - int(int(self.font_size[0] * 1.3))
            self.spr.set_label_attributes(int(self.font_size[0] + 0.5),
                                          True, 'right', y_pos=y, i=0)
        elif self.name in block_styles['clamp-style-boolean'] or \
                self.name in block_styles['clamp-style-until']:
            y = self.docks[1][3] - int(int(self.font_size[0] * 1.3))
            self.spr.set_label_attributes(int(self.font_size[0] + 0.5),
                                          True, 'right', y_pos=y, i=0)
            y = self.docks[2][3] - int(int(self.font_size[0] * 1.9))
            self.spr.set_label_attributes(int(self.font_size[1] + 0.5),
                                          True, 'right', y_pos=y, i=1)
        elif self.name in block_styles['clamp-style-else']:
            self.spr.set_margins(l=10 * self.scale, r=10 * self.scale)
            y = self.docks[1][3] - int(int(self.font_size[0] * 1.3))
            self.spr.set_label_attributes(int(self.font_size[0] + 0.5),
                                          True, 'right', y_pos=y, i=0)
            y = self.docks[2][3] - int(int(self.font_size[0] * 1.9))
            self.spr.set_label_attributes(int(self.font_size[1] + 0.5),
                                          True, 'left', y_pos=y, i=1)
            y = self.docks[3][3] - int(int(self.font_size[0] * 1.45))
            self.spr.set_label_attributes(int(self.font_size[1] + 0.5),
                                          True, 'left', y_pos=y, i=2)

    def _set_labels(self, i, label):
        if self.spr is None:
            return
        self.spr.set_label(label, i)

    def _make_block(self, svg):
        self._left = 0
        self._top = 0
        self._right = 0
        self._bottom = 0
        self.svg.set_stroke_width(STANDARD_STROKE_WIDTH)
        for k in list(block_styles.keys()):
            if self.name in block_styles[k]:
                if isinstance(self._block_methods[k], list):
                    self._block_methods[k][0](svg, self._block_methods[k][1],
                                              self._block_methods[k][2])
                else:
                    self._block_methods[k](svg)
                return
        error_output('ERROR: block type not found %s' % (self.name))
        self._block_methods['blank-style'](svg)
        self.unknown = True

    def _set_colors(self, svg):
        if self._custom_colors:
            self.svg.set_colors(self.colors)
            return
        if self.name in BOX_COLORS:
            self.colors = BOX_COLORS[self.name]
        elif self.name in special_block_colors:
            self.colors = special_block_colors[self.name]
        else:
            for p in range(len(palette_blocks)):
                if self.name in palette_blocks[p]:
                    self.colors = block_colors[p]
        self.svg.set_colors(self.colors)

    def _make_basic_style(self, svg, extend_x=0, extend_y=0):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow',
                       True,
                       self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow',
                       False,
                       self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_blank_style(self, svg, extend_x=0, extend_y=0):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(False)
        self.svg.set_tab(False)
        self.svg.set_tail(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = []

    def _make_basic_style_head(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_slot(False)
        self.svg.set_cap(True)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['unavailable', False, 0, 0],
                      ['flow', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]]]

    def _make_basic_style_head_1arg(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_innie([True])
        self.svg.set_slot(False)
        self.svg.set_cap(True)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['unavailable', False, 0, 0],
                      ['string', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_basic_style_tail(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_tab(False)
        self.svg.set_tail(True)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['unavailable', False, 0, 0]]

    def _make_basic_style_1arg(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_innie([True])
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['flow', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]]]

    def _make_basic_style_2arg(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_innie([True, True])
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['number', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['flow', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]]]

    def _make_basic_style_3arg(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey, 0, self.ey2)
        self.svg.set_innie([True, True, True])
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['number', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['number', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]],
                      ['flow', False, self.svg.docks[4][0],
                       self.svg.docks[4][1]]]

    def _make_basic_style_7arg(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_innie([True, True, True, True, True, True, True])
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['string', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['media', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['number', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]],
                      ['number', False, self.svg.docks[4][0],
                       self.svg.docks[4][1]],
                      ['number', False, self.svg.docks[5][0],
                       self.svg.docks[5][1]],
                      ['number', False, self.svg.docks[6][0],
                       self.svg.docks[6][1]],
                      ['number', False, self.svg.docks[7][0],
                       self.svg.docks[7][1]],
                      ['flow', False, self.svg.docks[8][0],
                       self.svg.docks[8][1]]]

    def _make_basic_style_var_arg(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        innie = [True]
        for i in range(self._ei):
            innie.append(True)
        self.svg.set_innie(innie)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]
        for i in range(self._ei):
            self.docks.append(['number', False, self.svg.docks[i + 2][0],
                               self.svg.docks[i + 2][1]])
        self.docks.append(['flow', False, self.svg.docks[self._ei + 2][0],
                           self.svg.docks[self._ei + 2][1]])

    def _make_bullet_style(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        innie = [True, True]
        for i in range(self._ei):
            innie.append(True)
        self.svg.set_innie(innie)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['string', False, self.svg.docks[1][0],
                       self.svg.docks[1][1], '['],
                      ['string', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]]]
        for i in range(self._ei):
            self.docks.append(['string', False, self.svg.docks[i + 3][0],
                               self.svg.docks[i + 3][1]])
        self.docks.append(['flow', False, self.svg.docks[self._ei + 3][0],
                           self.svg.docks[self._ei + 3][1], ']'])

    def _make_box_style(self, svg):
        self.svg.expand(60 + self.dx + self.ex, self.ey)
        self._make_block_graphics(svg, self.svg.basic_box)
        self.docks = [['number', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['unavailable', False, 0, 0]]

    def _make_media_style(self, svg):
        self.svg.expand(40 + self.dx + self.ex, 10 + self.ey)
        self._make_block_graphics(svg, self.svg.basic_box)
        self.docks = [['number', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['unavailable', False, 0, 0]]

    def _make_number_style(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        self.svg.set_innie([True, True])
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        """
        NOTE: The "outie" is added last, so the dock order in NUMBER_STYLE
              blocks needs to be modified.
        """
        self.docks = [['number', True, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_number_style_var_arg(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        innie = [True]
        for i in range(self._ei + 1):
            innie.append(True)
        self.svg.set_innie(innie)
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['number', True, self.svg.docks[2 + self._ei][0],
                       self.svg.docks[2 + self._ei][1]],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]]]
        for i in range(self._ei + 1):
            self.docks.append(['number', False, self.svg.docks[i + 1][0],
                               self.svg.docks[i + 1][1]])
        self.docks.append(['unavailable', False, 0, 0])

    def _make_number_style_var_3arg(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        _ei = 1
        innie = [True]
        for i in range(_ei + 1):
            innie.append(True)
        self.svg.set_innie(innie)
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['number', True, self.svg.docks[2 + _ei][0],
                       self.svg.docks[2 + _ei][1]],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]]]

        for i in range(_ei + 1):
            self.docks.append(['number', False, self.svg.docks[i + 1][0],
                               self.svg.docks[i + 1][1]])
        self.docks.append(['unavailable', False, 0, 0])

    def _make_number_style_block(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        self.svg.set_innie([True, True])
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['number', True, self.svg.docks[2][0],
                       self.svg.docks[2][1], '('],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['unavailable', False, 0, 0, ')']]

    def _make_number_style_1arg(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        self.svg.set_innie([True])
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['number', True, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]]]

    def _make_number_style_1strarg(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        self.svg.set_innie([True])
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['number', True, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['string', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['unavailable', False, 0, 0]]

    def _make_number_style_porch(self, svg):
        self.svg.expand(self.dx + self.ex, self.ey)
        self.svg.set_innie([True, True])
        self.svg.set_outie(True)
        self.svg.set_tab(False)
        self.svg.set_slot(False)
        self.svg.set_porch(True)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['number', True, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_compare_style(self, svg):
        self.svg.expand(15 + self.dx + self.ex, self.ey)
        self._make_block_graphics(svg, self.svg.boolean_compare)
        self.docks = [['bool', True, self.svg.docks[0][0],
                       self.svg.docks[0][1], '('],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['number', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['unavailable', False, 0, 0, ')']]

    def _make_compare_porch_style(self, svg):
        self.svg.set_porch(True)
        self._make_compare_style(svg)

    def _make_boolean_style(self, svg):
        self.svg.expand(15 + self.dx + self.ex, self.ey)
        self._make_block_graphics(svg, self.svg.boolean_and_or)
        self.docks = [['bool', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['bool', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['bool', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]]]

    def _make_not_style(self, svg):
        self.svg.expand(15 + self.dx + self.ex, self.ey)
        self._make_block_graphics(svg, self.svg.boolean_not, arg=False)
        self.docks = [['bool', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['bool', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_boolean_block_style(self, svg):
        self.svg.expand(15 + self.dx + self.ex, self.ey)
        self._make_block_graphics(svg, self.svg.boolean_not, arg=True)
        self.docks = [['bool', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['unavailable', False, 0, 0]]

    def _make_boolean_1arg_block_style(self, svg):
        self.svg.expand(15 + self.dx + self.ex, self.ey)
        self.svg.set_innie([True])
        self._make_block_graphics(svg, self.svg.boolean_not, arg=True)
        self.docks = [['bool', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_clamp_style(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1], '['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[3][0],
                       self.svg.docks[3][1], ']']]

    def _make_clamp_style_collapsible(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_collapsible(True)
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1], '['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[3][0],
                       self.svg.docks[3][1], ']']]

    def _make_clamp_style_collapsed(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_show(True)
        self._make_block_graphics(svg, self.svg.basic_block)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow', True, 0, self.svg.docks[0][1] + 10, '['],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1], ']']]

    def _make_clamp_style_1arg(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_innie([True])
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['number', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['flow', False, self.svg.docks[2][0],
                       self.svg.docks[2][1], '['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[4][0],
                       self.svg.docks[4][1], ']']]

    def _make_clamp_style_hat(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(False)
        self.svg.set_cap(True)
        self.svg.set_tab(True)
        self.svg.set_tail(True)
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['unavailable', False, 0, 0],
                      ['flow', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]]]

    def _make_clamp_style_hat_1arg(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(False)
        self.svg.set_cap(True)
        self.svg.set_tab(True)
        self.svg.set_tail(True)
        self.svg.set_innie([True])
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['unavailable', False, 0, 0],
                      ['number', False, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_clamp_style_boolean(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_boolean(True)
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['bool', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['flow', False, self.svg.docks[2][0],
                       self.svg.docks[2][1], '['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[4][0],
                       self.svg.docks[4][1], ']']]

    def _make_clamp_style_until(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y,
                        0, self.ey2)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_boolean(True)
        self.svg.second_clamp(False)
        self._make_block_graphics(svg, self.svg.clamp_until)
        # Dock positions are flipped
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['bool', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1], '['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[4][0],
                       self.svg.docks[4][1], ']']]

    def _make_clamp_style_else(self, svg, extend_x=0, extend_y=4):
        self.svg.expand(self.dx + self.ex + extend_x, self.ey + extend_y,
                        self.dx + self.ex + extend_x, self.ey2 + extend_y)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_boolean(True)
        self.svg.second_clamp(True)
        self._make_block_graphics(svg, self.svg.clamp)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['bool', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['flow', False, self.svg.docks[2][0],
                       self.svg.docks[2][1], '['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[4][0],
                       self.svg.docks[4][1], ']['],
                      # Skip bottom of clamp
                      ['flow', False, self.svg.docks[6][0],
                       self.svg.docks[6][1], ']']]

    def _make_flow_style_tail(self, svg):
        self.svg.expand(10 + self.dx + self.ex, self.ey)
        self.svg.set_slot(True)
        self.svg.set_tab(False)
        self._make_block_graphics(svg, self.svg.basic_flow)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    # Depreciated block styles

    def _make_portfolio_style_2x2(self, svg):
        self.svg.expand(30 + self.dx + self.ex, 10 + self.ey)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_innie([True, True, False, True])
        self._make_block_graphics(svg, self.svg.portfolio)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['string', False, self.svg.docks[6][0],
                       self.svg.docks[6][1]],
                      ['media', False, self.svg.docks[5][0],
                       self.svg.docks[5][1]],
                      ['media', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['media', False, self.svg.docks[4][0],
                       self.svg.docks[4][1]],
                      ['media', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['flow', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]]]

    def _make_portfolio_style_2x1(self, svg):
        self.svg.expand(30 + self.dx + self.ex, 10 + self.ey)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_innie([True, True])
        self._make_block_graphics(svg, self.svg.portfolio)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['string', False, self.svg.docks[4][0],
                       self.svg.docks[4][1]],
                      ['media', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]],
                      ['media', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]],
                      ['flow', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]]]

    def _make_portfolio_style_1x2(self, svg):
        self.svg.expand(30 + self.dx + self.ex, 15 + self.ey)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_innie([True, True, False, True])
        self.svg.set_draw_innies(False)
        self._make_block_graphics(svg, self.svg.portfolio)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['string', False, self.svg.docks[4][0],
                       self.svg.docks[4][1]],
                      ['media', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]],
                      ['media', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_portfolio_style_1x1(self, svg):
        self.svg.expand(30 + self.dx + self.ex, 15 + self.ey)
        self.svg.set_slot(True)
        self.svg.set_tab(True)
        self.svg.set_innie([True, True])
        self.svg.set_draw_innies(False)
        self._make_block_graphics(svg, self.svg.portfolio)
        self.docks = [['flow', True, self.svg.docks[0][0],
                       self.svg.docks[0][1]],
                      ['string', False, self.svg.docks[3][0],
                       self.svg.docks[3][1]],
                      ['media', False, self.svg.docks[2][0],
                       self.svg.docks[2][1]],
                      ['flow', False, self.svg.docks[1][0],
                       self.svg.docks[1][1]]]

    def _make_block_graphics(self, svg, function, arg=None):
        self._set_colors(svg)
        self.svg.set_gradient(True, GRADIENT_COLOR)
        self.svg.clear_docks()
        if arg is None:
            pixbuf = svg_str_to_pixbuf(function())
        else:
            pixbuf = svg_str_to_pixbuf(function(arg))
        self.width = self.svg.get_width()
        self.height = self.svg.get_height()
        self.shapes[0] = _pixbuf_to_cairo_surface(pixbuf,
                                                  self.width, self.height)
        self.svg.set_gradient(False)
        self.svg.clear_docks()
        if arg is None:
            pixbuf = svg_str_to_pixbuf(function())
        else:
            pixbuf = svg_str_to_pixbuf(function(arg))
        self.shapes[1] = _pixbuf_to_cairo_surface(pixbuf,
                                                  self.width, self.height)


def _pixbuf_to_cairo_surface(image, width, height):
    surface = cairo.ImageSurface(
        cairo.FORMAT_ARGB32, int(width), int(height))
    context = cairo.Context(surface)
    Gdk.cairo_set_source_pixbuf(context, image, 0, 0)
    context.rectangle(0, 0, int(width), int(height))
    context.fill()
    return surface



# Copyright (C) 2010 Collabora Ltd. <http://www.collabora.co.uk/>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import logging

import dbus
from dbus import PROPERTIES_IFACE

from telepathy.interfaces import CLIENT, \
    CLIENT_APPROVER, \
    CLIENT_HANDLER, \
    CLIENT_INTERFACE_REQUESTS
from telepathy.server import DBusProperties

from . import dispatch


SUGAR_CLIENT_SERVICE = 'org.freedesktop.Telepathy.Client.Sugar'
SUGAR_CLIENT_PATH = '/org/freedesktop/Telepathy/Client/Sugar'

_instance = None


class TelepathyClient(dbus.service.Object, DBusProperties):

    def __init__(self):
        self._interfaces = set([CLIENT, CLIENT_HANDLER,
                                CLIENT_INTERFACE_REQUESTS, PROPERTIES_IFACE,
                                CLIENT_APPROVER])

        bus = dbus.Bus()
        bus_name = dbus.service.BusName(SUGAR_CLIENT_SERVICE, bus=bus)

        dbus.service.Object.__init__(self, bus_name, SUGAR_CLIENT_PATH)
        DBusProperties.__init__(self)

        self._implement_property_get(CLIENT, {
            'Interfaces': lambda: list(self._interfaces), })
        self._implement_property_get(CLIENT_HANDLER, {
            'HandlerChannelFilter': self.__get_filters_cb, })
        self._implement_property_get(CLIENT_APPROVER, {
            'ApproverChannelFilter': self.__get_filters_cb, })

        self.got_channel = dispatch.Signal()
        self.got_dispatch_operation = dispatch.Signal()

    def __get_filters_cb(self):
        logging.debug('__get_filters_cb')
        filter_dict = dbus.Dictionary({}, signature='sv')
        return dbus.Array([filter_dict], signature='a{sv}')

    @dbus.service.method(dbus_interface=CLIENT_HANDLER,
                         in_signature='ooa(oa{sv})aota{sv}', out_signature='')
    def HandleChannels(self, account, connection, channels, requests_satisfied,
                       user_action_time, handler_info):
        logging.debug('HandleChannels\n%r\n%r\n%r\n%r\n%r\n%r\n', account,
                      connection, channels, requests_satisfied,
                      user_action_time, handler_info)
        for channel in channels:
            self.got_channel.send(self, account=account,
                                  connection=connection, channel=channel)

    @dbus.service.method(dbus_interface=CLIENT_INTERFACE_REQUESTS,
                         in_signature='oa{sv}', out_signature='')
    def AddRequest(self, request, properties):
        logging.debug('AddRequest\n%r\n%r', request, properties)

    @dbus.service.method(dbus_interface=CLIENT_APPROVER,
                         in_signature='a(oa{sv})oa{sv}', out_signature='',
                         async_callbacks=('success_cb', 'error_cb_'))
    def AddDispatchOperation(self, channels, dispatch_operation_path,
                             properties, success_cb, error_cb_):
        success_cb()
        try:
            logging.debug('AddDispatchOperation\n%r\n%r\n%r', channels,
                          dispatch_operation_path, properties)

            self.got_dispatch_operation.send(
                self,
                channels=channels,
                dispatch_operation_path=dispatch_operation_path,
                properties=properties)
        except Exception as e:
            logging.exception(e)


def get_instance():
    global _instance
    if not _instance:
        _instance = TelepathyClient()
    return _instance



# -*- coding: utf-8 -*-
# Copyright (C) 2006-2010 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#
TOOLSVERSION = u"ODFPY/0.9.6"

ANIMNS = u"urn:oasis:names:tc:opendocument:xmlns:animation:1.0"
CHARTNS = u"urn:oasis:names:tc:opendocument:xmlns:chart:1.0"
CHARTOOONS = u"http://openoffice.org/2010/chart"
CONFIGNS = u"urn:oasis:names:tc:opendocument:xmlns:config:1.0"
CSS3TNS = u"http://www.w3.org/TR/css3-text/"
#DBNS           = u"http://openoffice.org/2004/database"
DBNS = u"urn:oasis:names:tc:opendocument:xmlns:database:1.0"
DCNS = u"http://purl.org/dc/elements/1.1/"
DOMNS = u"http://www.w3.org/2001/xml-events"
DR3DNS = u"urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"
DRAWNS = u"urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"
FIELDNS = u"urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0"
FONS = u"urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"
FORMNS = u"urn:oasis:names:tc:opendocument:xmlns:form:1.0"
FORMXNS = u"urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:form:1.0"
GRDDLNS = u"http://www.w3.org/2003/g/data-view#"
KOFFICENS = u"http://www.koffice.org/2005/"
MANIFESTNS = u"urn:oasis:names:tc:opendocument:xmlns:manifest:1.0"
MATHNS = u"http://www.w3.org/1998/Math/MathML"
METANS = u"urn:oasis:names:tc:opendocument:xmlns:meta:1.0"
NUMBERNS = u"urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"
OFFICENS = u"urn:oasis:names:tc:opendocument:xmlns:office:1.0"
OFNS = u"urn:oasis:names:tc:opendocument:xmlns:of:1.2"
OOOCNS = u"http://openoffice.org/2004/calc"
OOONS = u"http://openoffice.org/2004/office"
OOOWNS = u"http://openoffice.org/2004/writer"
PRESENTATIONNS = u"urn:oasis:names:tc:opendocument:xmlns:presentation:1.0"
RDFANS = u"http://docs.oasis-open.org/opendocument/meta/rdfa#"
RPTNS = u"http://openoffice.org/2005/report"
SCRIPTNS = u"urn:oasis:names:tc:opendocument:xmlns:script:1.0"
SMILNS = u"urn:oasis:names:tc:opendocument:xmlns:smil-compatible:1.0"
STYLENS = u"urn:oasis:names:tc:opendocument:xmlns:style:1.0"
SVGNS = u"urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"
TABLENS = u"urn:oasis:names:tc:opendocument:xmlns:table:1.0"
TABLEOOONS = u"http://openoffice.org/2009/table"
TEXTNS = u"urn:oasis:names:tc:opendocument:xmlns:text:1.0"
XFORMSNS = u"http://www.w3.org/2002/xforms"
XHTMLNS = u"http://www.w3.org/1999/xhtml"
XLINKNS = u"http://www.w3.org/1999/xlink"
XMLNS = u"http://www.w3.org/XML/1998/namespace"
XSDNS = u"http://www.w3.org/2001/XMLSchema"
XSINS = u"http://www.w3.org/2001/XMLSchema-instance"

nsdict = {
    ANIMNS: u'anim',
    CHARTNS: u'chart',
    CHARTOOONS: u'chartooo',
    CONFIGNS: u'config',
    CSS3TNS: u'css3t',
    DBNS: u'db',
    DCNS: u'dc',
    DOMNS: u'dom',
    DR3DNS: u'dr3d',
    DRAWNS: u'draw',
    FIELDNS: u'field',
    FONS: u'fo',
    FORMNS: u'form',
    FORMXNS: u'formx',
    GRDDLNS: u'grddl',
    KOFFICENS: u'koffice',
    MANIFESTNS: u'manifest',
    MATHNS: u'math',
    METANS: u'meta',
    NUMBERNS: u'number',
    OFFICENS: u'office',
    OFNS: u'of',
    OOONS: u'ooo',
    OOOWNS: u'ooow',
    OOOCNS: u'oooc',
    PRESENTATIONNS: u'presentation',
    RDFANS: u'rdfa',
    RPTNS: u'rpt',
    SCRIPTNS: u'script',
    SMILNS: u'smil',
    STYLENS: u'style',
    SVGNS: u'svg',
    TABLENS: u'table',
    TABLEOOONS: u'tableooo',
    TEXTNS: u'text',
    XFORMSNS: u'xforms',
    XLINKNS: u'xlink',
    XHTMLNS: u'xhtml',
    XMLNS: u'xml',
    XSDNS: u'xsd',
    XSINS: u'xsi',
}



# Copyright (c) 2008-13, Walter Bender

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from sugar3.datastore import datastore
from TurtleArt.tapalette import (logo_commands, logo_functions)
from TurtleArt.taconstants import (TITLEXY, CONSTANTS)


def save_logo(tw):
    """ Set up the Turtle Art color palette and color processing. """

    # We need to catch several special cases: stacks, boxes, labels, etc.
    dispatch_table = {
        'label': _add_label,
        'to action': _add_named_stack,
        'action': _add_reference_to_stack,
        'storeinbox': _add_named_box,
        'box': _add_reference_to_box}
    constants_table = {
        'lpos': _lpos,
        'tpos': _tpos,
        'rpos': _rpos,
        'bpos': _bpos,
        'red': _red,
        'orange': _orange,
        'yellow': _yellow,
        'green': _green,
        'cyan': _cyan,
        'blue': _blue,
        'purple': _purple,
        'white': _white,
        'black': _black,
        'titlex': _titlex,
        'titley': _titley,
        'leftx': _leftx,
        'topy': _topy,
        'rightx': _rightx,
        'bottomy': _bottomy,
        'width': _width,
        'height': _height}

    stacks_of_blocks = tw.just_blocks()
    stack_count = 0

    logocode = ''

    """
    Walk through the code, substituting UCB Logo for Turtle Art primitives.
    """
    for stack in stacks_of_blocks:
        this_stack = ''
        psuedocode = _walk_stack(tw, stack)
        if psuedocode == []:
            continue

        skip = False
        for i in range(len(psuedocode)):
            if skip:
                skip = False
                continue
            blk = psuedocode[i]
            if isinstance(blk, tuple):
                (blk, _blk_no) = blk
            if blk in logo_commands:
                logo_command = logo_commands[blk]
            else:
                logo_command = None
            if i == 0 and logo_command not in ['to stack1\n', 'to stack2\n',
                                               'to action', 'to start\n']:
                this_stack = 'to turtleblocks_%d\n' % (stack_count)
                stack_count += 1
            if logo_command in dispatch_table:
                if i + 1 < len(psuedocode):
                    this_stack += dispatch_table[logo_command](
                        psuedocode[i + 1])
                    skip = True
                else:
                    print('missing arg to %s' % (logo_command))
            elif logo_command in constants_table:
                this_stack += str(constants_table[logo_command](tw))
            elif logo_command is not None:
                this_stack += logo_command
            else:  # assume it is an argument
                if blk not in ['nop', 'nop1', 'nop2', 'nop3']:
                    if isinstance(blk, str) and blk[0:2] == '#s':
                        this_stack += str(blk[2:]).replace(' ', '_')
                    else:
                        this_stack += str(blk).replace(' ', '_')
            this_stack += ' '

        logocode += this_stack
        logocode += '\nend\n'

    # We may need to prepend some additional procedures.
    for key in logo_functions.keys():
        if key in logocode:
            logocode = logo_functions[key] + logocode

    if 'tasetshade' in logocode or 'tasetpencolor' in logocode or \
       'tasetbackground' in logocode:
        logocode = logo_functions['tacolor'] + logocode

    logocode = 'window\n' + logocode
    return logocode


def _add_label(string):
    if isinstance(string, str) and string[0:8] in ['#smedia_', '#saudio_',
                                                   '#svideo_', '#sdescr_']:
        string = string[8:]
        dsobject = datastore.get(string[8:])
        if 'title' in dsobject.metadata:
            string = dsobject.metadata['title']
    else:
        string = str(string)
    if string[0:2] == '#s':
        string = string[2:]
        string = '"' + string
    if string.count(' ') > 0:
        return 'label sentence %s\n' % (string.replace(' ', ' "'))
    else:
        return 'label %s' % (string.replace(' ', '_'))


def _add_named_stack(action):
    if isinstance(action, str) and action[0:2] == '#s':
        return 'to %s\n' % (str(action[2:]).replace(' ', '_'))
    else:
        return 'to %s\n' % (str(action).replace(' ', '_'))


def _add_reference_to_stack(action):
    if isinstance(action, str) and action[0:2] == '#s':
        return '%s' % (str(action[2:]).replace(' ', '_'))
    else:
        return '%s' % (str(action).replace(' ', '_'))


def _add_named_box(box_name):
    if isinstance(box_name, str) and box_name[0:2] == '#s':
        return 'make "%s' % (str(box_name[2:]).replace(' ', '_'))
    else:
        return 'make "%s' % (str(box_name).replace(' ', '_'))


def _add_reference_to_box(box_name):
    if isinstance(box_name, str) and box_name[0:2] == '#s':
        return ':%s' % (str(box_name[2:]).replace(' ', '_'))
    else:
        return ':%s' % (str(box_name).replace(' ', '_'))


def _lpos(tw):
    return int(-tw.canvas.width / (tw.coord_scale * 2))


def _tpos(tw):
    return int(tw.canvas.height / (tw.coord_scale * 2))


def _rpos(tw):
    return int(tw.canvas.width / (tw.coord_scale * 2))


def _bpos(tw):
    return int(-tw.canvas.height / (tw.coord_scale * 2))


def _width(tw):
    return int(tw.canvas.width / tw.coord_scale)


def _height(tw):
    int(tw.canvas.height / tw.coord_scale)


def _titlex(tw):
    return int(-(tw.canvas.width * TITLEXY[0]) / (tw.coord_scale * 2))


def _titley(tw):
    return int((tw.canvas.height * TITLEXY[1]) / (tw.coord_scale * 2))


def _leftx(tw):
    return int(-(tw.canvas.width * TITLEXY[0]) / (tw.coord_scale * 2))


def _topy(tw):
    return int((tw.canvas.height * (TITLEXY[1] - 0.125)) /
               (tw.coord_scale * 2))


def _rightx(tw):
    return 0


def _bottomy(tw):
    return 0


def _red(tw):
    return CONSTANTS['red']


def _orange(tw):
    return CONSTANTS['orange']


def _yellow(tw):
    return CONSTANTS['yellow']


def _green(tw):
    return CONSTANTS['green']


def _cyan(tw):
    return CONSTANTS['cyan']


def _blue(tw):
    return CONSTANTS['blue']


def _purple(tw):
    return CONSTANTS['purple']


def _white(tw):
    return 1


def _black(tw):
    return 0


def _walk_stack(tw, blk_in_stack):
    """ Convert blocks to logo psuedocode. """
    from .tautils import find_top_block

    top = find_top_block(blk_in_stack)
    if blk_in_stack == top:
        return tw.lc.generate_code(top, tw.just_blocks())
    else:
        return []



from device import RFIDDevice
from serial import Serial
import dbus
from dbus.mainloop.glib import DBusGMainLoop
import gobject
import re
from time import sleep

HAL_SERVICE = 'org.freedesktop.Hal'
HAL_MGR_PATH = '/org/freedesktop/Hal/Manager'
HAL_MGR_IFACE = 'org.freedesktop.Hal.Manager'
HAL_DEV_IFACE = 'org.freedesktop.Hal.Device'
REGEXP_SERUSB = '\/org\/freedesktop\/Hal\/devices\/usb_device['\
                'a-z,A-Z,0-9,_]*serial_usb_[0-9]'

STATE_WAITING = 0
STATE_WAITING2 = 1
STATE_READING = 2


class RFIDReader(RFIDDevice):
    """
    TIS-2000 interface.
    """

    def __init__(self):

        RFIDDevice.__init__(self)
        self.last_tag = ""
        self.ser = Serial()
        self.device = ''
        self.device_path = ''
        self._connected = False
        self._state = STATE_WAITING

        loop = DBusGMainLoop()
        self.bus = dbus.SystemBus(mainloop=loop)
        hmgr_iface = dbus.Interface(
            self.bus.get_object(
                HAL_SERVICE,
                HAL_MGR_PATH),
            HAL_MGR_IFACE)

        hmgr_iface.connect_to_signal('DeviceRemoved', self._device_removed_cb)

    def get_present(self):
        """
        Checks if TI-S2000 device is present.
        Returns True if so, False otherwise.
        """
        hmgr_if = dbus.Interface(
            self.bus.get_object(
                HAL_SERVICE,
                HAL_MGR_PATH),
            HAL_MGR_IFACE)
        tiusb_devices = set(hmgr_if.FindDeviceStringMatch('serial.type', 'usb')) & set(
            hmgr_if.FindDeviceStringMatch('info.product', 'TUSB3410 Microcontroller'))
        for i in tiusb_devices:
            tiusb_if = dbus.Interface(self.bus.get_object(HAL_SERVICE, i),
                                      HAL_DEV_IFACE)
            if tiusb_if.PropertyExists('linux.device_file'):
                self.device = str(tiusb_if.GetProperty('linux.device_file'))
                self.device_path = i
                return True
        return False

    def do_connect(self):
        """
        Connects to the device.
        Returns True if successfull, False otherwise.
        """
        retval = False
        if self.get_present():
            try:
                self.ser = Serial(self.device, 9600, timeout=0.1)
                self._connected = True
                self._escape()
                self._clear()
                self._format()
                gobject.idle_add(self._loop)
                retval = True
            except BaseException:
                self._connected = False
        return retval

    def do_disconnect(self):
        """
        Disconnect from the device.
        """
        self.ser.close()
        self._connected = False

    def read_tag(self):
        """
        Returns the last read value.
        """
        return self.last_tag

    def write_tag(self, hexval):
        """
        Usage: write_tag(hexval)

            Writes the hexadecimal string "hexval" into the tag.
            Returns True if successfull, False otherwise.
        """
        # self.ser.flushInput()
        reg = re.compile('([^0-9A-F]+)')
        if not (hexval.__len__() == 16 and reg.findall(hexval) == []):
            return False
        self.ser.read(100)
        self.ser.write('P')
        for i in hexval:
            self.ser.write(i)
        sleep(1)
        resp = self.ser.read(64)
        resp = resp.split()[0]
        if resp == "P0":
            return True
        else:
            return False

    def _escape(self):
        """
        Sends the scape command to the TIS-2000 device.
        """
        try:
            # self.ser.flushInput()
            self.ser.read(100)
            self.ser.write('\x1B')
            resp = self.ser.read()
            if resp == 'E':
                return True
            else:
                return False
        except BaseException:
            return False

    def _format(self):
        """
        Sends the format command to the TIS-2000 device.
        """
        try:
            # self.ser.flushInput()
            self.ser.read(100)
            self.ser.write('F')
            resp = self.ser.read()
            if resp == 'F':
                return True
            else:
                return False
        except BaseException:
            return False

    def _clear(self):
        """
        Sends the clear command to the TIS-2000 device.
        """
        try:
            # self.ser.flushInput()
            self.ser.read(100)
            self.ser.write('C')
            resp = self.ser.read()
            if resp == 'C':
                return True
            else:
                return False
        except BaseException:
            return False

    def get_version(self):
        """
        Sends the version command to the TIS-2000 device and returns
        a string with the device version.
        """
        # self.ser.flushInput()
        self.ser.read(100)
        self.ser.write('V')
        version = []
        tver = ""
        while True:
            resp = self.ser.read()
            if resp == '\x0A' or resp == '':
                break
            if resp != '\n' and resp != '\r':
                version.append(resp)
        for i in version:
            tver = tver + i
        if tver != "":
            return tver
        return "Unknown"

    def _device_removed_cb(self, path):
        """
        Called when a device is removed.
        Checks if the removed device is itself and emits the "disconnected"
        signal if so.
        """
        if path == self.device_path:
            self.device_path = ''
            self.ser.close()
            self._connected = False
            self.emit("disconnected", "TIS-2000")

    def _loop(self):
        """
        Threaded loop for reading data sent from the TIS-2000.
        """
        if not self._connected:
            return False

        if self._state is STATE_WAITING:
            data = self.ser.read()
            if data in ['W', 'R']:
                self._state = STATE_WAITING2
            return True

        elif self._state is STATE_WAITING2:
            data = self.ser.read()
            if data.isspace():
                self._state = STATE_READING
            else:
                self._clear()
                self._state = STATE_WAITING
            return True

        elif self._state is STATE_READING:
            data = self.ser.read(16)
            if data.__len__() < 16:
                self._clear()
                self._state = STATE_WAITING
            else:
                reg = re.compile('([^0-9A-F]+)')
                if reg.findall(data) == []:
                    self.emit("tag-read", data)
                    self.last_tag = data
                self._clear()
                self._state = STATE_WAITING
            return True
        return True

# Testing
# if __name__ == '__main__':
#    def handler(device, idhex):
#        """
#        Handler for "tag-read" signal.
#        Prints the tag id.
#        """
#        print "ID: ", idhex
#
#    dev = RFIDReader()
#    if dev.get_present():
#        dev.do_connect()
#        dev.connect('tag-read', handler)
#    else:
#        print "Not connected"
#
#    mloop = gobject.MainLoop()
#    mloop.run()



#!/usr/bin/env python
# Copyright (c) 2011, 2012 Walter Bender
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from gettext import gettext as _

try:
    from numpy.fft import rfft
    PITCH_AVAILABLE = True
except BaseException:
    PITCH_AVAILABLE = False

from plugins.plugin import Plugin

from plugins.audio_sensors.audiograb import (
    AudioGrab, SENSOR_DC_NO_BIAS, SENSOR_DC_BIAS, SENSOR_AC_BIAS)

from plugins.audio_sensors.ringbuffer import RingBuffer1d

from TurtleArt.tapalette import make_palette
from TurtleArt.taconstants import XO1, XO15, XO175, XO30, XO4
from TurtleArt.tautils import debug_output
from TurtleArt.taprimitive import (ConstantArg, Primitive)
from TurtleArt.tatype import TYPE_NUMBER

import logging
_logger = logging.getLogger('turtleart-activity audio sensors plugin')


def _avg(array, abs_value=False):
    ''' Calc. the average value of an array '''
    if len(array) == 0:
        return 0
    array_sum = 0
    if abs_value:
        for a in array:
            array_sum += abs(a)
    else:
        for a in array:
            array_sum += a
    return float(array_sum) / len(array)


class Audio_sensors(Plugin):

    def __init__(self, parent):
        Plugin.__init__(self)
        self._parent = parent
        self.audio_started = False
        self._sound_init = False
        self._resistance_init = False
        self._voltage_init = False
        self._status = True  # TODO: test for audio device
        # These flags are referenced by audiograb
        self.hw = self._parent.hw
        self.running_sugar = self._parent.running_sugar

    def setup(self):
        ''' set up audio-sensor-specific blocks '''
        self._sound = [0, 0]
        self._volume = [0, 0]
        self._pitch = [0, 0]
        self._resistance = [0, 0]
        self._voltage = [0, 0]
        self.max_samples = 1500
        self.input_step = 1
        self.ringbuffer = []

        palette = make_palette('sensor',
                               colors=["#FF6060", "#A06060"],
                               help_string=_('Palette of sensor blocks'),
                               position=6)
        hidden = True
        if self._status:
            hidden = False

        palette.add_block('sound',
                          hidden=hidden,
                          style='box-style',
                          label=_('sound'),
                          help_string=_('raw microphone input signal'),
                          value_block=True,
                          prim_name='sound')
        palette.add_block('volume',
                          hidden=hidden,
                          style='box-style',
                          label=_('loudness'),
                          help_string=_('microphone input volume'),
                          value_block=True,
                          prim_name='volume')

        self._parent.lc.def_prim(
            'sound', 0,
            Primitive(self.prim_sound,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(0)},
                      call_afterwards=self.after_sound))

        self._parent.lc.def_prim(
            'volume', 0,
            Primitive(self.prim_volume,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(0)},
                      call_afterwards=self.after_volume))

        hidden = True
        if PITCH_AVAILABLE and self._status:
            hidden = False

        palette.add_block('pitch',
                          hidden=hidden,
                          style='box-style',
                          label=_('pitch'),
                          help_string=_('microphone input pitch'),
                          value_block=True,
                          prim_name='pitch')
        self._parent.lc.def_prim(
            'pitch', 0,
            Primitive(self.prim_pitch,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(0)},
                      call_afterwards=self.after_pitch))

        hidden = True
        if self.hw in [XO1, XO15, XO175, XO4, XO30] and self._status:
            # Calibration based on http://bugs.sugarlabs.org/ticket/4649
            if self.hw == XO1:
                self.voltage_gain = 0.000022
                self.voltage_bias = 1.14
            elif self.hw == XO15:
                self.voltage_gain = -0.00015
                self.voltage_bias = 1.70
            elif self.hw == XO175:  # Range 0.01V to 3.01V
                self.voltage_gain = 0.0000516
                self.voltage_bias = 1.3598
            elif self.hw == XO4:  # Range 0.17V to 3.08V
                self.voltage_gain = 0.0004073
                self.voltage_bias = 1.6289
            else:  # XO 3.0
                self.voltage_gain = 0.000077
                self.voltage_bias = 0.72
            hidden = False

        palette.add_block('resistance',
                          hidden=hidden,
                          style='box-style',
                          label=_('resistance'),
                          help_string=_('microphone input resistance'),
                          prim_name='resistance')
        palette.add_block('voltage',
                          hidden=hidden,
                          style='box-style',
                          label=_('voltage'),
                          help_string=_('microphone input voltage'),
                          prim_name='voltage')

        hidden = True
        # Only add stereo capture for XO15 (broken on ARM #3675)
        if self.hw in [XO15] and self._status:
            hidden = False

        palette.add_block('resistance2',
                          hidden=hidden,
                          style='box-style',
                          label=_('resistance') + '2',
                          help_string=_('microphone input resistance'),
                          prim_name='resistance2')
        palette.add_block('voltage2',
                          hidden=hidden,
                          style='box-style',
                          label=_('voltage') + '2',
                          help_string=_('microphone input voltage'),
                          prim_name='voltage2')
        self._parent.lc.def_prim(
            'resistance', 0,
            Primitive(self.prim_resistance,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(0)},
                      call_afterwards=self.after_resistance))
        self._parent.lc.def_prim(
            'voltage', 0,
            Primitive(self.prim_voltage,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(0)},
                      call_afterwards=self.after_voltage))
        self._parent.lc.def_prim(
            'resistance2', 0,
            Primitive(self.prim_resistance,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(1)},
                      call_afterwards=self.after_resistance))
        self._parent.lc.def_prim(
            'voltage2', 0,
            Primitive(self.prim_voltage,
                      return_type=TYPE_NUMBER,
                      kwarg_descs={'channel': ConstantArg(1)},
                      call_afterwards=self.after_voltage))

        if self.hw in [XO15, XO175, XO30, XO4]:
            self.PARAMETERS = {
                SENSOR_AC_BIAS: (False, True, 80, True),
                SENSOR_DC_NO_BIAS: (True, False, 80, False),
                SENSOR_DC_BIAS: (True, True, 90, False)
            }
        elif self.hw == XO1:
            self.PARAMETERS = {
                SENSOR_AC_BIAS: (False, True, 40, True),
                SENSOR_DC_NO_BIAS: (True, False, 0, False),
                SENSOR_DC_BIAS: (True, True, 0, False)
            }
        else:
            self.PARAMETERS = {
                SENSOR_AC_BIAS: (None, True, 40, True),
                SENSOR_DC_NO_BIAS: (True, False, 80, False),
                SENSOR_DC_BIAS: (True, True, 90, False)
            }

    def start(self):
        ''' Start grabbing audio if there is an audio block in use '''
        if not self._status:
            return

        self._sound_init = False
        self._resistance_init = False
        self._voltage_init = False

        self._sound = [0, 0]
        self._volume = [0, 0]
        self._pitch = [0, 0]
        self._resistance = [0, 0]
        self._voltage = [0, 0]

        if self.audio_started:
            self.audiograb.stop_grabbing()

    def new_buffer(self, buf, channel=0):
        ''' Append a new buffer to the ringbuffer '''
        self.ringbuffer[channel].append(buf)
        return True

    def stop(self):
        ''' This gets called by the stop button '''
        if self._status and self.audio_started:
            self.audiograb.on_activity_quit()  # reset all setting
        self.audio_started = False

    def goto_background(self):
        ''' This gets called when your process is sent to the background '''
        pass

    def return_to_foreground(self):
        ''' This gets called when your process returns from the background '''
        pass

    def quit(self):
        ''' This gets called by the quit button '''
        if self._status and self.audio_started:
            self.audiograb.on_activity_quit()

    def _status_report(self):
        debug_output(
            'Reporting audio sensor status: %s' % (str(self._status)),
            self._parent.running_sugar)
        return self._status

    # Block primitives

    def prim_sound(self, channel=0):
        if not self._status:
            return 0

        self._init_sound()

        self._prim_sound(0)
        # Return average of both channels if sampling in stereo
        if self._channels == 2:
            self._prim_sound(1)
            return (self._sound[0] + self._sound[1]) / 2.0
        else:
            return self._sound[0]

    def _init_sound(self):
        if not self._sound_init:
            mode, bias, gain, boost = self.PARAMETERS[SENSOR_AC_BIAS]
            self.audiograb = AudioGrab(self.new_buffer, self,
                                       mode, bias, gain, boost)
            self._channels = self.audiograb.channels
            for i in range(self._channels):
                self.ringbuffer.append(RingBuffer1d(self.max_samples,
                                                    dtype='int16'))
            self.audiograb.start_grabbing()
            self.audio_started = True
            self._sound_init = True

    def _prim_sound(self, channel):
        ''' return raw mic in value '''
        buf = self.ringbuffer[channel].read(None, self.input_step)
        if len(buf) > 0:
            self._sound[channel] = float(buf[0])
        else:
            self._sound[channel] = 0

    def after_sound(self, channel=0):
        if self._parent.lc.update_values:
            self._parent.lc.update_label_value('sound', self._sound[channel])

    def prim_volume(self, channel=0):
        if not self._status:
            return 0

        self._init_sound()

        self._prim_volume(0)
        # Return average of both channels if sampling in stereo
        if self._channels == 2:
            self._prim_volume(1)
            return (self._volume[0] + self._volume[1]) / 2.0
        else:
            return self._volume[0]

    def _prim_volume(self, channel):
        ''' return raw mic in value '''
        buf = self.ringbuffer[channel].read(None, self.input_step)
        if len(buf) > 0:
            self._volume[channel] = float(_avg(buf, abs_value=True))
        else:
            self._volume[channel] = 0

    def after_volume(self, channel=0):
        if self._parent.lc.update_values:
            self._parent.lc.update_label_value('volume', self._volume[channel])

    def prim_pitch(self, channel=0):
        if not self._status:
            return 0

        self._init_sound()

        self._prim_pitch(0)
        # Return average of both channels if sampling in stereo
        if self._channels == 2:
            self._prim_pitch(1)
            return (self._pitch[0] + self._pitch[1]) / 2.0
        else:
            return self._pitch[0]

    def _prim_pitch(self, channel):
        ''' return raw mic in value '''
        buf = self.ringbuffer[channel].read(None, self.input_step)
        if len(buf) > 0:
            buf = rfft(buf)
            buf = abs(buf)
            maxi = buf.argmax()
            if maxi == 0:
                self._pitch[channel] = 0
            else:  # Simple interpolation
                a, b, c = buf[maxi - 1], buf[maxi], buf[maxi + 1]
                maxi -= a / float(a + b + c)
                maxi += c / float(a + b + c)
                self._pitch[channel] = maxi * 48000 / (len(buf) * 2)
        else:
            self._pitch[channel] = 0

    def after_pitch(self, channel=0):
        if self._parent.lc.update_values:
            self._parent.lc.update_label_value('pitch', self._pitch[channel])

    def prim_resistance(self, channel=0):
        if self.hw not in [XO1, XO15, XO175, XO30, XO4] or not self._status:
            return 0

        if not self._resistance_init:
            mode, bias, gain, boost = self.PARAMETERS[SENSOR_DC_BIAS]
            self.audiograb = AudioGrab(self.new_buffer, self,
                                       mode, bias, gain, boost)
            self._channels = self.audiograb.channels
            for i in range(self._channels):
                self.ringbuffer.append(RingBuffer1d(self.max_samples,
                                                    dtype='int16'))
            self.audiograb.start_grabbing()
            self.audio_started = True
            self._resistance_init = True

        if self.hw in [XO1, XO4]:
            self._prim_resistance(0)
            return self._resistance[0]
        elif self.hw == XO15:
            self._prim_resistance(channel)
            return self._resistance[channel]
        # For XO175: channel assignment is seemingly random
        # (#3675), one of them will be 0
        else:
            self._prim_resistance(0)
            if self._resistance[0] != 999999999:
                return self._resistance[0]
            else:
                self._prim_resistance(1)
                return self._resistance[1]

    def _prim_resistance(self, channel):
        ''' return resistance sensor value '''
        buf = self.ringbuffer[channel].read(None, self.input_step)
        if len(buf) > 0:
            # See http://bugs.sugarlabs.org/ticket/552#comment:7
            # and http://bugs.sugarlabs.org/ticket/4649
            avg_buf = float(_avg(buf))
            if self.hw == XO1:
                self._resistance[channel] = \
                    2.718 ** ((avg_buf * 0.000045788) + 8.0531)
            elif self.hw == XO15:
                if avg_buf > 0:
                    self._resistance[channel] = (420000000 / avg_buf) - 13500
                else:
                    self._resistance[channel] = 420000000
            elif self.hw == XO175:  # Range 0 to inf ohms
                if avg_buf < 30519:
                    self._resistance[channel] = \
                        (92000000. / (30519 - avg_buf)) - 1620
                else:
                    self._resistance[channel] = 999999999
            elif self.hw == XO4:  # Range 0 to inf ohms
                if avg_buf < 6629:
                    self._resistance[channel] = \
                        (50000000. / (6629 - avg_buf)) - 3175
                else:
                    self._resistance[channel] = 999999999
            else:  # XO 3.0
                if avg_buf < 30514:
                    self._resistance[channel] = \
                        (46000000. / (30514 - avg_buf)) - 1150
                else:
                    self._resistance[channel] = 999999999
            if self._resistance[channel] < 0:
                self._resistance[channel] = 0
        else:
            self._resistance[channel] = 0

    def after_resistance(self, channel=0):
        if self._parent.lc.update_values:
            self._parent.lc.update_label_value(
                ['resistance', 'resistance2'][channel],
                self._resistance[channel])

    def prim_voltage(self, channel=0):
        if self.hw not in [XO1, XO15, XO175, XO30, XO4] or not self._status:
            return 0

        if not self._voltage_init:
            mode, bias, gain, boost = self.PARAMETERS[SENSOR_DC_NO_BIAS]
            self.audiograb = AudioGrab(self.new_buffer, self,
                                       mode, bias, gain, boost)
            self._channels = self.audiograb.channels
            for i in range(self._channels):
                self.ringbuffer.append(RingBuffer1d(self.max_samples,
                                                    dtype='int16'))
            self.audiograb.start_grabbing()
            self.audio_started = True
            self._voltage_init = True

        if self.hw in [XO1, XO4]:
            self._prim_voltage(0)
            return self._voltage[0]
        elif self.hw == XO15:
            self._prim_voltage(channel)
            return self._voltage[channel]
        # FIXME: For XO175: channel assignment is seemingly random
        # (#3675), one of them will be 0
        else:
            self._prim_voltage(0)
            if self._voltage[0] != 0:
                return self._voltage[0]
            else:
                self._prim_voltage(1)
                return self._voltage[1]

    def _prim_voltage(self, channel):
        ''' return voltage sensor value '''
        buf = self.ringbuffer[channel].read(None, self.input_step)
        buf = self.ringbuffer[channel].read(None, self.input_step)
        if len(buf) > 0:
            # See <http://bugs.sugarlabs.org/ticket/552#comment:7>
            self._voltage[channel] = \
                float(_avg(buf)) * self.voltage_gain + self.voltage_bias
        else:
            self._voltage[channel] = 0

    def after_voltage(self, channel=0):
        if self._parent.lc.update_values:
            self._parent.lc.update_label_value(
                ['voltage', 'voltage2'][channel],
                self._voltage[channel])



# -*- coding: utf-8 -*-
# Copyright (c) 2010, Walter Bender
# Copyright (c) 2010, Tony Forster

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from gi.repository import Gst
Gst.init(None)
import time
from TurtleArt.tautils import debug_output


class Camera():
    ''' Sets up a pipe from the camera to a pixbuf and emits a signal
    when the image is ready. '''

    def __init__(self, device='/dev/video0'):
        ''' Prepare camera pipeline to pixbuf and signal watch '''
        self.pipe = Gst.Pipeline()
        v4l2src = Gst.ElementFactory.make('v4l2src', None)
        v4l2src.props.device = device
        self.pipe.add(v4l2src)
        videoconvert = Gst.ElementFactory.make('videoconvert', None)
        self.pipe.add(videoconvert)
        self.gdkpixbufsink = Gst.ElementFactory.make('gdkpixbufsink', None)
        self.pipe.add(self.gdkpixbufsink)
        v4l2src.link(videoconvert)
        videoconvert.link(self.gdkpixbufsink)
        if self.pipe is not None:
            self.bus = self.pipe.get_bus()
            self.bus.add_signal_watch()
            self.bus.connect('message', self._on_message)

    def _on_message(self, bus, message):
        ''' We get a message if a pixbuf is available '''
        if message.get_structure() is not None:
            if message.get_structure().get_name() == 'pixbuf':
                self.pixbuf = self.gdkpixbufsink.get_property("last-pixbuf")
                self.image_ready = True

    def start_camera_input(self):
        ''' Start grabbing '''
        self.pixbuf = None
        self.image_ready = False
        self.pipe.set_state(Gst.State.PLAYING)
        while not self.image_ready:
            self.bus.poll(Gst.MessageType.ANY, 1)

    def stop_camera_input(self):
        ''' Stop grabbing '''
        self.pipe.set_state(Gst.State.NULL)



# copyright (c) 2007-8, Playful Invention Company
# Copyright (c) 2008-13, Walter Bender
# Copyright (c) 2013 Alan Aguiar

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import tempfile
import dbus
import cairo
import pickle
import subprocess
import os
import string
import mimetypes
from gettext import gettext as _
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import GdkPixbuf
from gi.repository import Gio
import json
json.dumps
from json import load as jload
from json import dump as jdump
from io import StringIO

from .taconstants import (HIT_HIDE, HIT_SHOW, XO1, XO15, XO175, XO4, UNKNOWN,
                         MAGICNUMBER, SUFFIX, ARG_MUST_BE_NUMBER)

import logging
_logger = logging.getLogger('turtleart-activity')


FIRST_TIME = True


def debug_output(message_string, running_sugar=False):
    ''' unified debugging output '''
    if running_sugar:
        _logger.debug(message_string)
    else:
        print(message_string)


def error_output(message_string, running_sugar=False):
    ''' unified debugging output '''
    if running_sugar:
        _logger.error(message_string)
    else:
        print(message_string)


class pythonerror(Exception):

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


def convert(x, fn, try_ord=True):
    '''
    The strategy for mixing numbers and strings is to first try
    converting the string to a float; then if the string is a single
    character, try converting it to an ord; finally, just treat it as a
    string. Numbers appended to strings are first treated as ints, then
    floats.
    '''
    try:
        return fn(x)
    except ValueError:
        if try_ord:
            xx, flag = chr_to_ord(x)
            if flag:
                return fn(xx)
        return x


def chr_to_ord(x):
    ''' Try to comvert a string to an ord '''
    if strtype(x) and len(x) == 1:
        try:
            return ord(x), True
        except ValueError:
            return x, False
    return x, False


def strtype(x):
    ''' Is x a string type? '''
    return isinstance(x, str)


def increment_name(name):
    ''' If name is of the form foo_1, change it to foo_2. Otherwise,
    return name_1'''
    if '_' in name:
        parts = name.split('_')
        try:
            i = int(parts[-1])
            i += 1
            parts[-1] = str(i)
            newname = string.join(parts, '_')
        except ValueError:
            newname = '%s_1' % (name)
    else:
        newname = '%s_1' % (name)
    return newname


def magnitude(pos):
    ''' Calculate the magnitude of the distance between to blocks. '''
    x, y = pos
    return x * x + y * y


def json_load(text):
    ''' Load JSON data using what ever resources are available. '''

    # Remove MAGIC NUMBER, if present, and leading whitespace
    if text[0:2] == MAGICNUMBER:
        clean_text = text[2:].lstrip()
    else:
        clean_text = text.lstrip()
    # Strip out trailing whitespace, nulls, and newlines
    clean_text = clean_text.replace('\12', '')
    clean_text = clean_text.replace('\00', '')
    clean_text = clean_text.rstrip()
    # Look for missing ']'s
    left_count = clean_text.count('[')
    right_count = clean_text.count(']')
    while left_count > right_count:
        clean_text += ']'
        right_count = clean_text.count(']')
    io = StringIO(clean_text)
    try:
        listdata = jload(io)
    except ValueError:
        # Assume that text is ascii list
        listdata = text.split()
        for i, value in enumerate(listdata):
            listdata[i] = convert(value, float)
    # json converts tuples to lists, so we need to convert back,
    return _tuplify(listdata)


def find_hat(data):
    ''' Find a hat in a stack '''
    for i, blk in enumerate(data):
        if isinstance(blk[1], (tuple, list)) and _to_str(blk[1][0]) == 'hat':
            return i
    return None


def _to_str(text):
    ''' Convert whatever to a str type '''
    if isinstance(text, str):
        return text
    else:
        try:
            return str(text)
        except ValueError:
            return ''


def hat_on_top(data):
    ''' Move the hat block in a stack to the top '''
    hat = find_hat(data)
    if hat is None or hat == 0:
        return data

    data_was_tuple = False
    if isinstance(data, tuple):
        data_was_tuple = True
        data = listify(data)

    # First put the vertical flow together
    stack = [hat]
    sort = [hat]
    sorted_data = [[0, 'hat', data[hat][2], data[hat][3], []]]
    i = 1
    while len(stack) > 0:
        branch = stack[0]
        stack.remove(stack[0])
        if branch is None:
            continue
        else:
            blk = data[branch]

        if branch != hat:
            sort.append(blk[0])
            sorted_data.append([i, blk[1], 0, 0, []])
            i += 1

        while blk[4][-1] is not None:
            cblk = blk[4][-1]
            sort.append(cblk)
            name = data[cblk][1]
            sorted_data.append([i, name, 0, 0, []])
            if isinstance(name, list):
                name = name[0]
            # Some blocks have multiple branches
            if _to_str(name) in ['repeat', 'forever', 'while', 'until',
                                 'ifelse', 'if', 'sandwichclamp',
                                 'sandwichclampcollapsed']:
                stack.append(data[cblk][4][-2])
            if _to_str(name) in ['ifelse']:
                stack.append(data[cblk][4][-3])
            i += 1
            blk = data[cblk]

    # Then add the argument blocks
    for blk in data:
        if blk[0] in sort:
            continue
        sort.append(blk[0])
        sorted_data.append([i, blk[1], 0, 0, []])
        i += 1

    # Then add the connections
    for i, blk in enumerate(sort):
        for j in data[blk][4]:
            if j is None:
                sorted_data[i][4].append(None)
            else:
                sorted_data[i][4].append(sort.index(j))

    if data_was_tuple:
        return _tuplify(sorted_data)
    return sorted_data


def _tuplify(tup):
    ''' Convert to tuples '''
    if not isinstance(tup, list):
        return tup
    return tuple(map(_tuplify, tup))


def listify(tup):
    ''' Convert to list '''
    if not isinstance(tup, tuple):
        return tup
    return list(map(listify, tup))


def get_id(connection):
    ''' Get a connection block ID. '''
    if connection is not None and hasattr(connection, 'id'):
        return connection.id
    return None


def json_dump(data):
    ''' Save data using available JSON tools. '''
    io = StringIO()
    jdump(data, io)
    return io.getvalue()


def get_endswith_files(path, end):
    f = os.listdir(path)
    files = []
    for name in f:
        if name.endswith(end):
            files.append(os.path.join(path, name))
    return files


def get_load_name(filefilter, load_save_folder=None):
    ''' Open a load file dialog. '''
    dialog = Gtk.FileChooserDialog(
        _('Load...'), None,
        Gtk.FileChooserAction.OPEN, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                     Gtk.STOCK_OPEN, Gtk.ResponseType.OK))
    dialog.set_default_response(Gtk.ResponseType.OK)
    return do_dialog(dialog, filefilter, load_save_folder)


def get_save_name(filefilter, load_save_folder, save_file_name):
    ''' Open a save file dialog. '''
    dialog = Gtk.FileChooserDialog(
        _('Save...'), None,
        Gtk.FileChooserAction.SAVE, (Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
                                     Gtk.STOCK_SAVE, Gtk.ResponseType.OK))
    dialog.set_default_response(Gtk.ResponseType.OK)
    if filefilter in ['.png', '.svg', '.lg', '.py', '.odp']:
        suffix = filefilter
    else:
        suffix = SUFFIX[1]
    if save_file_name is not None:
        if not save_file_name.endswith(suffix):
            save_file_name = save_file_name + suffix
        dialog.set_current_name(save_file_name)
    return do_dialog(dialog, filefilter, load_save_folder)


def chooser_dialog(parent_window, filter, action):
    ''' Choose an object from the datastore and take some action '''
    from sugar3.graphics.objectchooser import ObjectChooser

    chooser = None
    dsobject = None
    cleanup_needed = False
    try:
        chooser = ObjectChooser(parent=parent_window, what_filter=filter)
    except TypeError:  # Old-syle Sugar chooser
        chooser = ObjectChooser(
            None,
            parent_window,
            Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT)
        cleanup_needed = True

    if chooser is not None:
        result = chooser.run()
        if result == Gtk.ResponseType.ACCEPT:
            dsobject = chooser.get_selected_object()
        if cleanup_needed:
            chooser.destroy()
            del chooser
    GLib.idle_add(action, dsobject)


def data_from_file(ta_file):
    ''' Open the .ta file, ignoring any .png file that might be present. '''
    file_handle = open(ta_file, 'r')
    #
    # We try to maintain read-compatibility with all versions of Turtle Art.
    # Try pickle first; then different versions of json.
    #
    try:
        data = pickle.load(file_handle)
    except BaseException:
        # Rewind necessary because of failed pickle.load attempt
        file_handle.seek(0)
        text = file_handle.read()
        data = data_from_string(text)
    file_handle.close()
    return data


def data_from_string(text):
    ''' JSON load data from a string. '''
    if isinstance(text, str):
        return json_load(text.replace(']],\n', ']], '))
    else:
        print('type error (%s) in data_from_string' % (type(text)))
        return None


def data_to_file(data, ta_file):
    ''' Write data to a file. '''
    try:
        file_handle = open(ta_file, 'w')
    except IOError as e:
        error_output('Could not write to %s: %s.' % (ta_file, e))
        tmp_file = os.path.join(os.path.expanduser('~'),
                                os.path.basename(ta_file))
        try:
            debug_output('Trying to write to %s' % (tmp_file))
            file_handle = open(tmp_file, 'w')
        except IOError as e:
            error_output('Could not write to %s: %s.' % (tmp_file, e))
            tmp_file = os.path.join(tempfile.gettempdir(),
                                    os.path.basename(ta_file))
            try:
                debug_output('Trying to write to %s' % (tmp_file))
                file_handle = open(tmp_file, 'w')
            except IOError as e:
                error_output('Could not write to %s: %s.' % (tmp_file, e))
                return
    file_handle.write(data_to_string(data))
    file_handle.close()


def data_to_string(data):
    ''' JSON dump a string. '''
    return json_dump(data).replace(']], ', ']],\n')


def do_dialog(dialog, suffix, load_save_folder):
    ''' Open a file dialog. '''
    result = None
    file_filter = Gtk.FileFilter()
    file_filter.add_pattern('*' + suffix)
    file_filter.set_name('Turtle Art')
    dialog.add_filter(file_filter)

    if load_save_folder is not None:
        dialog.set_current_folder(load_save_folder)

    response = dialog.run()
    if response == Gtk.ResponseType.OK:
        result = dialog.get_filename()
        load_save_folder = dialog.get_current_folder()
    dialog.destroy()
    return result, load_save_folder


def save_picture(canvas, file_name):
    ''' Save the canvas to a file '''
    x_surface = canvas.canvas.get_target()
    img_surface = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                     canvas.width, canvas.height)
    cr = cairo.Context(img_surface)
    cr.set_source_surface(x_surface)
    cr.paint()
    img_surface.write_to_png(str(file_name))


def get_canvas_data(canvas):
    ''' Get pixel data from the turtle canvas '''
    x_surface = canvas.canvas.get_target()
    img_surface = cairo.ImageSurface(cairo.FORMAT_RGB24,
                                     canvas.width, canvas.height)
    cr = cairo.Context(img_surface)
    cr.set_source_surface(x_surface)
    cr.paint()
    return img_surface.get_data()


def get_pixbuf_from_journal(dsobject, w, h):
    ''' Load a pixbuf from a Journal object. '''
    if hasattr(dsobject, 'file_path'):
        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(dsobject.file_path,
                                                        int(w), int(h))
    else:
        pixbufloader = GdkPixbuf.PixbufLoader.new_with_mime_type('image/png')
        pixbufloader.set_size(min(300, int(w)), min(225, int(h)))
        try:
            pixbufloader.write(dsobject.metadata['preview'])
        except Exception as e:
            error_output('could not read preview: %s' % e, True)
            return None
        pixbufloader.close()
        pixbuf = pixbufloader.get_pixbuf()

    return pixbuf


def get_path(activity, subpath):
    ''' Find a Rainbow-approved place for temporary files. '''
    if hasattr(activity, "get_activity_root"):
        return(os.path.join(activity.get_activity_root(), subpath))
    else:
        # Early versions of Sugar didn't support get_activity_root()
        return(os.path.join(os.environ['HOME'], '.sugar/default',
                            'org.laptop.TurtleArtActivity', subpath))


def image_to_base64(image_path, tmp_path):
    ''' Convert an image to base64-encoded data '''
    base64 = os.path.join(tmp_path, 'base64tmp')
    cmd = 'base64 <' + image_path + ' >' + base64
    subprocess.check_call(cmd, shell=True)
    file_handle = open(base64, 'r')
    data = file_handle.read()
    file_handle.close()
    os.remove(base64)
    return data


def base64_to_image(data, path_name):
    ''' Convert base64-encoded data to an image '''
    base64 = os.path.join(path_name, 'base64tmp')
    file_handle = open(base64, 'w')
    file_handle.write(data)
    file_handle.close()
    file_name = os.path.join(path_name, 'imagetmp.png')
    cmd = 'base64 -d <' + base64 + '>' + file_name
    subprocess.check_call(cmd, shell=True)
    return file_name


def movie_media_type(name):
    ''' Is it movie media? '''
    guess = mimetypes.guess_type(name)
    if guess[0] is None:
        return False
    return guess[0][0:5] == 'video'


def audio_media_type(name):
    ''' Is it audio media? '''
    guess = mimetypes.guess_type(name)
    if guess[0] is None:
        return False
    return guess[0][0:5] == 'audio'


def image_media_type(name):
    ''' Is it image media? '''
    guess = mimetypes.guess_type(name)
    if guess[0] is None:
        return False
    return guess[0][0:5] == 'image'


def text_media_type(name):
    ''' Is it text media? '''
    guess = mimetypes.guess_type(name)
    if guess[0] is None:
        return False
    return guess[0][0:4] == 'text'


def round_int(num):
    ''' Remove trailing decimal places if number is an int '''
    try:
        float(num)
    except TypeError:
        raise pythonerror('#syntaxerror')

    if int(float(num)) == num:
        return int(num)
    else:
        if float(num) < 0:
            nn = int((float(num) - 0.005) * 100) / 100.
        else:
            nn = int((float(num) + 0.005) * 100) / 100.
        if int(float(nn)) == nn:
            return int(nn)
        return nn


def calc_image_size(spr):
    ''' Calculate the maximum size for placing an image onto a sprite. '''
    return int(max(spr.label_safe_width(), 1)), \
        int(max(spr.label_safe_height(), 1))


def restore_clamp(top):
    ''' Restore the blocks in a sandwich clamp. '''
    if top is None:
        return

    if top.name == 'sandwichclampcollapsed':
        y1 = top.docks[2][3]
        if top.connections[1] is not None:
            for blk in find_group(top.connections[1]):
                blk.spr.restore()
                blk.status = None

            # If you come across a 'sandwichclampcollapsed', do not
            # restore its clamp
            for blk in find_group(top.connections[1]):
                if blk.name == 'sandwichclampcollapsed':
                    if blk.connections[1] is not None:
                        for b in find_group(blk.connections[1]):
                            b.spr.hide()
                            b.status = 'collapsed'

        bot = top.connections[2]
        top.name = 'sandwichclamp'
        top.spr.set_label('')
        top.resize()
        top.svg.set_hide(True)
        top.refresh()
        y2 = top.docks[2][3]
        dy = y2 - y1
        if bot is not None:
            for blk in find_group(bot):
                blk.spr.move_relative((0, dy))
        if top.connections[1] is not None:
            # Make sure stack is aligned to dock
            x1, y1 = top.connections[1].spr.get_xy()
            x1 += top.connections[1].docks[0][2]
            y1 += top.connections[1].docks[0][3]
            x2, y2 = top.spr.get_xy()
            x2 += top.docks[1][2]
            y2 += top.docks[1][3]
            if x1 != x2 or y1 != y2:
                for blk in find_group(top.connections[1]):
                    blk.spr.move_relative((x2 - x1, y2 - y1))
        return


def collapse_clamp(top, transform=True):
    ''' Hide all the blocks in the clamp. '''
    if top is None or top.spr is None:
        return

    if top.name in ['sandwichclampcollapsed', 'sandwichclamp']:
        y1 = top.docks[2][3]
        if top.connections[1] is not None:
            for blk in find_group(top.connections[1]):
                blk.spr.hide()
                blk.status = 'collapsed'
        if transform:
            bot = top.connections[2]
            top.name = 'sandwichclampcollapsed'
            top.spr.set_label(_('click to open'))
            top.reset_y()
            top.resize()
            top.svg.set_hide(False)
            top.refresh()
            y2 = top.docks[2][3]
            dy = y2 - y1
            if bot is not None:
                for blk in find_group(bot):
                    blk.spr.move_relative((0, dy))
        return


def hide_button_hit(spr, x, y):
    ''' Did the sprite's hide (contract) button get hit? '''
    red, green, blue, alpha = spr.get_pixel((x, y))
    if red == HIT_HIDE:
        return True
    else:
        return False


def show_button_hit(spr, x, y):
    ''' Did the sprite's show (expand) button get hit? '''
    red, green, blue, alpha = spr.get_pixel((x, y))
    if green == HIT_SHOW:
        return True
    else:
        return False


def numeric_arg(value):
    ''' Dock test: looking for a numeric value '''
    if isinstance(convert(value, float), float):
        return True
    return False


def zero_arg(value):
    ''' Dock test: looking for a zero argument '''
    if numeric_arg(value):
        if convert(value, float) == 0:
            return True
    return False


def neg_arg(value):
    ''' Dock test: looking for a negative argument '''
    if numeric_arg(value):
        if convert(value, float) < 0:
            return True
    return False


def journal_check(blk1, blk2, dock1, dock2):
    ''' Dock blocks only if arg is Journal block '''
    if blk1 is None or blk2 is None:
        return True
    if (blk1.name == 'skin' and dock1 == 1) and blk2.name != 'journal':
        return False
    if (blk2.name == 'skin' and dock2 == 1) and blk1.name != 'journal':
        return False
    return True


def arithmetic_check(blk1, blk2, dock1, dock2):
    ''' Dock strings only if they convert to numbers. Avoid /0 and root(-1)'''
    if blk1 is None or blk2 is None:
        return True
    if blk1.name in ['sqrt', 'number', 'string'] and\
       blk2.name in ['sqrt', 'number', 'string']:
        if blk1.name == 'number' or blk1.name == 'string':
            if not numeric_arg(blk1.values[0]) or neg_arg(blk1.values[0]):
                return False
        elif blk2.name == 'number' or blk2.name == 'string':
            if not numeric_arg(blk2.values[0]) or neg_arg(blk2.values[0]):
                return False
    elif blk1.name in ['division2', 'number', 'string'] and \
            blk2.name in ['division2', 'number', 'string']:
        if blk1.name == 'number' or blk1.name == 'string':
            if not numeric_arg(blk1.values[0]):
                return False
            if dock2 == 2 and zero_arg(blk1.values[0]):
                return False
        elif blk2.name == 'number' or blk2.name == 'string':
            if not numeric_arg(blk2.values[0]):
                return False
            if dock1 == 2 and zero_arg(blk2.values[0]):
                return False
    elif blk1.name in ARG_MUST_BE_NUMBER and blk2.name in ARG_MUST_BE_NUMBER:
        if blk1.name == 'string':
            if not numeric_arg(blk1.values[0]):
                return False
        elif blk2.name == 'string':
            if not numeric_arg(blk2.values[0]):
                return False
    elif blk1.name in ['greater2', 'less2'] and blk2.name == 'string':
        # Non-numeric stings are OK only if both args are strings;
        # Lots of test conditions...
        if dock1 == 1 and blk1.connections[2] is not None:
            if blk1.connections[2].name == 'number':
                if not numeric_arg(blk2.values[0]):
                    return False
        elif dock1 == 2 and blk1.connections[1] is not None:
            if blk1.connections[1].name == 'number':
                if not numeric_arg(blk2.values[0]):
                    return False
    elif blk2.name in ['greater2', 'less2'] and blk1.name == 'string':
        if dock2 == 1 and blk2.connections[2] is not None:
            if blk2.connections[2].name == 'number':
                if not numeric_arg(blk1.values[0]):
                    return False
        elif dock2 == 2 and blk2.connections[1] is not None:
            if blk2.connections[1].name == 'number':
                if not numeric_arg(blk1.values[0]):
                    return False
    elif blk1.name in ['greater2', 'less2'] and blk2.name == 'number':
        if dock1 == 1 and blk1.connections[2] is not None:
            if blk1.connections[2].name == 'string':
                if not numeric_arg(blk1.connections[2].values[0]):
                    return False
        elif dock1 == 2 and blk1.connections[1] is not None:
            if blk1.connections[1].name == 'string':
                if not numeric_arg(blk1.connections[1].values[0]):
                    return False
    elif blk2.name in ['greater2', 'less2'] and blk1.name == 'number':
        if dock2 == 1 and blk2.connections[2] is not None:
            if blk2.connections[2].name == 'string':
                if not numeric_arg(blk2.connections[2].values[0]):
                    return False
        elif dock2 == 2 and blk2.connections[1] is not None:
            if blk2.connections[1].name == 'string':
                if not numeric_arg(blk2.connections[1].values[0]):
                    return False
    return True


def xy(event):
    ''' Where is the mouse event? '''
    return list(map(int, event.get_coords()))


# Utilities related to finding blocks in stacks.


def find_block_to_run(blk):
    ''' Find a stack to run (any stack without a 'def action'on the top). '''
    top = find_top_block(blk)
    if blk == top and blk.name[0:3] is not 'def':
        return True
    else:
        return False


def find_top_block(blk):
    ''' Find the top block in a stack. '''
    if blk is None:
        return None
    if blk.connections is None:
        return blk
    if len(blk.connections) == 0:
        return blk
    while blk.connections[0] is not None:
        blk = blk.connections[0]
    return blk


def find_bot_block(blk):
    ''' Find the bottom block in a stack. '''
    if blk is None:
        return None
    if blk.connections is None:
        return blk
    if len(blk.connections) == 0:
        return blk
    while blk.connections[-1] is not None:
        blk = blk.connections[-1]
    return blk


def find_start_stack(blk):
    ''' Find a stack with a 'start' block on top. '''
    if blk is None:
        return False
    if find_top_block(blk).name == 'start':
        return True
    else:
        return False


def find_group(blk):
    ''' Find the connected group of block in a stack. '''
    if blk is None:
        return []
    group = [blk]
    if blk.connections is not None:
        for cblk in blk.connections[1:]:
            if cblk is not None:
                group.extend(find_group(cblk))
    return group


def find_blk_below(blk, namelist):
    ''' Find a specific block below this block. '''
    if blk is None or len(blk.connections) == 0:
        return
    if not isinstance(namelist, list):
        namelist = [namelist]
    group = find_group(blk)
    for gblk in group:
        if gblk.name in namelist:
            return gblk
    return None


def get_stack_width_and_height(blk):
    ''' What are the width and height of a stack? '''
    minx = 10000
    miny = 10000
    maxx = -10000
    maxy = -10000
    for gblk in find_group(blk):
        (x, y) = gblk.spr.get_xy()
        w, h = gblk.spr.get_dimensions()
        if x < minx:
            minx = x
        if y < miny:
            miny = y
        if x + w > maxx:
            maxx = x + w
        if y + h > maxy:
            maxy = y + h
    return(maxx - minx, maxy - miny)


def get_stack_name(blk):
    ''' Return the name of the action stack that the given block belongs to.
    If the top block of this stack is not a stack-defining block, return
    None. '''
    top_block = find_top_block(blk)
    if top_block.name == 'start':
        return 'start'
    elif top_block.name == 'hat1':
        return 'stack1'
    elif top_block.name == 'hat2':
        return 'stack2'
    elif top_block.name == 'hat':
        try:
            return str(top_block.connections[1].values[0])
        except (AttributeError, TypeError, IndexError):
            # AttributeError: t_b has no attribute 'connections' or t_b.c[1]
            #   has no attribute 'value'
            # TypeError: t_b.c or t_b.c[1].v is not a subscriptable sequence
            # IndexError: t_b.c or t_b.c[1].v is too short
            return None
    else:
        return None


def get_hardware():
    ''' Determine whether we are using XO 1.0, 1.5, ... or 'unknown'
    hardware '''
    version = _get_dmi('product_version')
    # product = _get_dmi('product_name')
    if version is None:
        hwinfo_path = '/bin/olpc-hwinfo'
        if os.path.exists(hwinfo_path) and os.access(hwinfo_path, os.X_OK):
            model = check_output([hwinfo_path, 'model'], 'unknown hardware')
            version = model.strip()
    if version == '1':
        return XO1
    elif version == '1.5':
        return XO15
    elif version == '1.75':
        return XO175
    elif version == '4':
        return XO4
    else:
        # Some systems (e.g. ARM) don't have dmi info
        if os.path.exists('/sys/devices/platform/lis3lv02d/position'):
            return XO175
        elif os.path.exists('/etc/olpc-release'):
            return XO1
        else:
            return UNKNOWN


def _get_dmi(node):
    ''' The desktop management interface should be a reliable source
    for product and version information. '''
    path = os.path.join('/sys/class/dmi/id', node)
    try:
        return open(path).readline().strip()
    except BaseException:
        return None


def get_screen_dpi():
    ''' Return screen DPI '''
    try:
        xft_dpi = Gtk.settings_get_default().get_property('gtk-xft-dpi')
        dpi = float(xft_dpi / 1024)
        return dpi
    except BaseException:
        return 100


def check_output(command, warning):
    ''' Workaround for old systems without subprocess.check_output'''
    if hasattr(subprocess, 'check_output'):
        try:
            output = subprocess.check_output(command)
        except subprocess.CalledProcessError:
            print(warning)
            return None
    else:
        import subprocess

        cmd = ''
        for c in command:
            cmd += c
            cmd += ' '
        (status, output) = subprocess.getstatusoutput(cmd)
        if status != 0:
            print(warning)
            return None
    return output


def power_manager_off(status):
    '''
    Power management in Sugar
         power_manager_off(True) --> Disable power manager
         power_manager_off(False) --> Use custom power manager
    '''

    global FIRST_TIME

    OHM_SERVICE_NAME = 'org.freedesktop.ohm'
    OHM_SERVICE_PATH = '/org/freedesktop/ohm/Keystore'
    OHM_SERVICE_IFACE = 'org.freedesktop.ohm.Keystore'
    PATH = '/etc/powerd/flags/inhibit-suspend'

    settings = Gio.Settings('org.sugarlabs.power')

    ACTUAL_POWER = True

    if FIRST_TIME:
        ACTUAL_POWER = settings.get_boolean('automatic')
        FIRST_TIME = False

    if status:
        VALUE = False
    else:
        VALUE = ACTUAL_POWER

    settings.set_bool('automatic', VALUE)

    bus = dbus.SystemBus()
    try:
        proxy = bus.get_object(OHM_SERVICE_NAME, OHM_SERVICE_PATH)
        keystore = dbus.Interface(proxy, OHM_SERVICE_IFACE)
        keystore.SetKey('suspend.automatic_pm', bool(VALUE))
    except dbus.exceptions.DBusException:
        if status:
            try:
                fd = open(PATH, 'w')
                fd.close()
            except IOError:
                pass
        elif ACTUAL_POWER:
            try:
                os.remove(PATH)
            except OSError:
                pass


def is_writeable(path):
    ''' Make sure we can write to the directory or file '''
    return os.access(path, os.W_OK)
    """
    if not os.path.exists(path):
        return False
    stats = os.stat(path)
    if (stats.st_uid == os.geteuid() and stats.st_mode & stat.S_IWUSR) or \
       (stats.st_gid == os.getegid() and stats.st_mode & stat.S_IWGRP) or \
       (stats.st_mode & stat.S_IWOTH):
        return True
    return False
    """
